from .db import Database
from .pose import Pose
from .cset import IngredientSet

import mcol

from typing import Callable

import os

import mrich


class PoseTable:
    """Class representing all :class:`.Pose` objects in the 'pose' table of the :class:`.Database`.

    .. attention::

            :class:`.PoseTable` objects should not be created directly. Instead use the :meth:`.HIPPO.poses` property. See :doc:`getting_started` and :doc:`insert_elaborations`.

    Use as an iterable
    ==================

    Iterate through :class:`.Pose` objects in the table:

    ::

            for pose in animal.poses:
                ...


    Selecting poses in the table
    ============================

    The :class:`.PoseTable` can be indexed with :class:`.Pose` IDs, names, aliases, or list/sets/tuples/slices thereof:

    ::

            ptable = animal.poses

            # indexing individual compounds
            pose = ptable[13]                            # using the ID
            pose = ptable["BSYNRYMUTXBXSQ-UHFFFAOYSA-N"] # using the InChIKey
            pose = ptable["Ax0310a"]                     # using the alias

            # getting a subset of compounds
            pset = ptable[13,15,18]      # using IDs (tuple)
            pset = ptable[[13,15,18]]    # using IDs (list)
            pset = ptable[set(13,15,18)] # using IDs (set)
            pset = ptable[13:18]         # using a slice

    Tags and target IDs can also be used to filter:

    ::

            pset = animal.poses(tag='hits') # select compounds tagged with 'hits'
            pset = animal.poses(target=1)   # select poses from the first target

    """

    _table = "pose"
    _name = "all poses"

    def __init__(
        self,
        db: Database,
    ) -> None:

        self._db = db
        self._interactions = None

    ### FACTORIES

    ### PROPERTIES

    @property
    def db(self) -> Database:
        """Returns the associated :class:`.Database`"""
        return self._db

    @property
    def table(self) -> str:
        """Returns the name of the :class:`.Database` table"""
        return self._table

    @property
    def name(self) -> str | None:
        """Returns the name of set"""
        return self._name

    @property
    def names(self) -> list[str]:
        """Returns the aliases of child poses"""
        return [p.name for p in self]

    @property
    def aliases(self) -> list[str]:
        """Returns the aliases of child poses"""
        result = self.db.select(table=self.table, query="pose_alias", multiple=True)
        return [q for q, in result]

    @property
    def inchikeys(self) -> list[str]:
        """Returns the inchikeys of child poses"""
        result = self.db.select(table=self.table, query="pose_inchikey", multiple=True)
        return [q for q, in result]

    @property
    def ids(self) -> list[int]:
        """Returns the IDs of child poses"""
        result = self.db.select(table=self.table, query="pose_id", multiple=True)
        return [q for q, in result]

    @property
    def tags(self) -> set[str]:
        """Returns the set of unique tags present in this pose set"""
        values = self.db.select_where(
            table="tag",
            query="DISTINCT tag_name",
            key="tag_pose IS NOT NULL",
            multiple=True,
        )
        return set(v for v, in values)

    @property
    def num_fingerprinted(self) -> int:
        """Count the number of fingerprinted poses"""
        return self.db.count_where(
            table="pose",
            key="fingerprint",
            value=1,
        )

    @property
    def id_name_dict(self) -> dict[int, str]:
        """Return a dictionary mapping pose ID's to their name"""

        records = self.db.select(
            table=self.table, query="pose_id, pose_inchikey, pose_alias", multiple=True
        )

        lookup = {}
        for i, inchikey, alias in records:
            if alias:
                lookup[i] = alias
            else:
                lookup[i] = inchikey

        return lookup

    @property
    def interactions(self) -> "InteractionSet":
        """Get a :class:`.InteractionSet`"""
        if self._interactions is None:
            from .iset import InteractionSet

            self._interactions = InteractionSet.all(self.db)

        return self._interactions

    ### METHODS

    def get_by_tag(
        self,
        tag: str,
        inverse: bool = False,
    ) -> "PoseSet":
        """Get all child poses with a certain tag

        :param tag: tag to search for
        :param inverse: invert the selection
        :returns: a :class:`.PoseSet` of the subset

        """

        if not inverse:

            values = self.db.select_where(
                query="tag_pose", table="tag", key="name", value=tag, multiple=True
            )

        else:

            values = self.db.select_where(
                query="tag_pose", table="tag", key="name", value=tag, multiple=True
            )

            if not values:
                return self

            ids = [v for v, in values if v]

            values = self.db.select_where(
                query="pose_id",
                table="pose",
                key=f"pose_id NOT IN {self.str_ids}",
                multiple=True,
            )

        if not values:
            return None

        ids = [v for v, in values if v]

        pset = self[ids]

        if inverse:
            pset._name = f'poses not tagged "{tag}"'
        else:
            pset._name = f'poses tagged "{tag}"'
        return pset

    def get_by_target(
        self,
        *,
        id: int,
    ) -> "PoseSet":
        """Get all child poses with a certain :class:`.Target` ID:

        :param id: :class:`.Target` ID
        :returns: a :class:`.PoseSet` of the subset

        """
        assert isinstance(id, int)
        values = self.db.select_where(
            query="pose_id", table="pose", key="target", value=id, multiple=True
        )
        ids = [v for v, in values if v]

        target = self.db.get_target(id=id)

        pset = self[ids]
        pset._name = f'poses for "{target}"'
        return pset

    def get_by_smiles(self, smiles: str) -> "Pose | PoseSet | None":
        """Get a member pose by it's smiles"""

        from .tools import inchikey_from_smiles, sanitise_smiles, SanitisationError

        try:
            flat_smiles = sanitise_smiles(smiles, sanitisation_failed="error")
        except SanitisationError as e:
            mrich.error(f"Could not sanitise {smiles=}")
            mrich.error(str(e))
            return None
        except AssertionError:
            mrich.error(f"Could not sanitise {smiles=}")
            return None
            return c

        # get the compound

        flat_inchikey = inchikey_from_smiles(flat_smiles)

        comp_id = self.db.select_id_where(
            table="compound", key="inchikey", value=flat_inchikey
        )

        if not comp_id:
            return None

        (comp_id,) = comp_id

        # get the poses

        pose_ids = self.db.select_id_where(
            table="pose", key="compound", value=comp_id, multiple=True
        )

        if not pose_ids:
            return None

        pose_ids = [i for i, in pose_ids]
        pset = self[pose_ids]

        # identify the pose

        inchikey = inchikey_from_smiles(smiles)

        matches = set()
        for pose in pset:
            if pose.inchikey == inchikey:
                matches.add(pose.id)
        matches = list(matches)

        if not matches:
            mrich.error(f"Did not find pose matching stereochemistry (C{comp_id})")
            return None

        if len(matches) == 1:
            return self[matches[0]]

        return self[matches]

    def get_by_subsite(
        self,
        *,
        id: int,
    ) -> "PoseSet":
        """Get all child poses with a certain :class:`.Subsite` ID:

        :param id: :class:`.Subsite` ID
        :returns: a :class:`.PoseSet` of the subset

        """
        assert isinstance(id, int)
        values = self.db.select_where(
            query="subsite_tag_pose",
            table="subsite_tag",
            key="ref",
            value=id,
            multiple=True,
        )
        ids = [v for v, in values if v]

        subsite = self.db.get_subsite_name(id=id)

        pset = self[ids]
        pset._name = f'poses in "{subsite}"'
        return pset

    def get_by_metadata(
        self,
        key: str,
        value: str | None = None,
    ) -> "PoseSet":
        """Get all child poses by their metadata. If no value is passed, then simply containing the key in the metadata dictionary is sufficient

        :param key: metadata key to match
        :param value: metadata value to match, if ``None`` any pose with the key present will be returned (Default value = None)
        :returns: a :class:`.PoseSet` of the subset

        """
        results = self.db.select(
            query="pose_id, pose_metadata", table="pose", multiple=True
        )
        if value is None:
            ids = [i for i, d in results if d and f'"{key}":' in d]
            name = f"poses with {key} in metadata"
        else:
            if isinstance(value, str):
                value = f'"{value}"'
            ids = [i for i, d in results if d and f'"{key}": {value}' in d]
            name = f"poses with metadata[{key}] == {value}"

        pset = self[ids]
        pset._name = name
        return pset

    def get_by_metadata_substring_match(
        self,
        substring: str,
    ) -> "PoseSet":
        """Get :class:`.PoseSet` of poses with metadata JSON containing substring"""

        assert substring
        assert isinstance(substring, str)

        pose_ids = self.db.select_where(
            table="pose",
            query="pose_id",
            key=f"""pose_metadata LIKE '%{substring}%'""",
            multiple=True,
        )

        if not pose_ids:
            mrich.error("No poses with export ")
            return None

        pose_ids = [i for i, in pose_ids]

        name = f"poses with '{substring}' in metadata"

        pset = self[pose_ids]
        pset._name = name

        return pset

    def draw(
        self,
        max_draw: int = 100,
    ) -> None:
        """Render the poses

        :param max_draw: show a warning if trying to draw more than this number of poses (Default value = 100)

        """
        if len(self) <= max_draw:
            self[:].draw()
        else:
            mrich.warning(
                f"Too many poses: {len(self)} > {max_draw=}. Increase max_draw or use animal.poses[:].draw()"
            )

    def summary(self) -> None:
        """Print a summary of this pose set"""
        mrich.header("PoseTable()")
        mrich.var("#poses", len(self))
        mrich.var("tags", self.tags)

    def interactive(self) -> None:
        """Interactive widget to navigate poses in the table

        .. attention::

                This method instantiates a :class:`.PoseSet` containing all poses, it is recommended to instead select a subset for display. This method is only intended for use within a Jupyter Notebook.

        """

        self[self.ids].interactive()

    ### DUNDERS

    def __call__(
        self,
        *,
        tag: str | None = None,
        target: int | None = None,
        subsite: int | None = None,
        smiles: str | None = None,
    ) -> "PoseSet":
        """Filter poses by a given tag, subsite ID, or target ID. See :meth:`.PoseTable.get_by_tag`, :meth:`.PoseTable.get_by_target`, amd :meth:`.PoseTable.get_by_subsite`"""

        if tag:
            return self.get_by_tag(tag)
        elif target:
            return self.get_by_target(id=target)
        elif subsite:
            return self.get_by_subsite(id=subsite)
        elif smiles:
            return self.get_by_smiles(smiles=smiles)
        else:
            raise NotImplementedError

    def __getitem__(
        self,
        key: int | str | tuple | list | set | slice,
    ) -> Pose:
        """Get a member :class:`.Pose` object or subset :class:`.PoseSet` thereof.

        :param key: Can be an integer ID, negative integer index, alias or inchikey string, list/set/tuple of IDs, or slice of IDs

        """

        from pandas import Series

        match key:

            case int():
                if key == 0:
                    return self.__getitem__(key=1)

                if key < 0:
                    key = len(self) + 1 + key
                    return self.__getitem__(key=key)

                else:
                    return self.db.get_pose(id=key)

            case str():
                pose = self.db.get_pose(alias=key)
                if not pose:
                    pose = self.db.get_pose(inchikey=key)
                return pose

            case key if (
                isinstance(key, list)
                or isinstance(key, tuple)
                or isinstance(key, set)
                or isinstance(key, Series)
            ):

                indices = []
                for i in key:
                    if isinstance(i, int):
                        index = i
                    elif isinstance(i, str):
                        index = self.db.get_pose_id(alias=i)
                        if not index:
                            index = self.db.get_pose_id(inchikey=i)
                    else:
                        raise NotImplementedError

                    assert index
                    indices.append(index)

                return PoseSet(self.db, indices)

            case slice():
                ids, name = self.db.slice_ids(
                    table=self.table,
                    start=key.start,
                    stop=key.stop,
                    step=key.step,
                    name=True,
                )
                pset = self[ids]
                pset._name = name
                return pset

            case _:
                mrich.error(
                    f"Unsupported type for PoseTable.__getitem__(): {type(key)}"
                )

        return None

    def __str__(self):
        """Unformatted string representation"""
        if self.name:
            s = f"{self.name}: "
        else:
            s = ""

        s += "{" f"P × {len(self)}" "}"

        return s

    def __repr__(self) -> str:
        """ANSI Formatted string representation"""
        return f"{mcol.bold}{mcol.underline}{self}{mcol.unbold}{mcol.ununderline}"

    def __rich__(self) -> str:
        """Rich Formatted string representation"""
        return f"[bold underline]{self}"

    def __len__(self) -> int:
        """Total number of compounds"""
        return self.db.count(self.table)

    def __iter__(self):
        """Iterate through all compounds"""
        return iter(self[i + 1] for i in range(len(self)))


class PoseSet:
    """Object representing a subset of the 'pose' table in the :class:`.Database`.

    .. attention::

            :class:`.PoseSet` objects should not be created directly. Instead use the :meth:`.HIPPO.poses` property. See :doc:`getting_started` and :doc:`insert_elaborations`.

    Use as an iterable
    ==================

    Iterate through :class:`.Pose` objects in the set:

    ::

            pset = animal.poses[:100]

            for pose in pset:
                    ...

    Check membership
    ================

    To determine if a :class:`.Pose` is present in the set:

    ::

            is_member = pose in cset

    Selecting compounds in the set
    ==============================

    The :class:`.PoseSet` can be indexed like standard Python lists by their indices

    ::

            pset = animal.poses[1:100]

            # indexing individual compounds
            pose = pset[0]  # get the first pose
            pose = pset[1]  # get the second pose
            pose = pset[-1] # get the last pose

            # getting a subset of compounds using a slice
            pset2 = pset[13:18] # using a slice

    """

    _table = "pose"

    def __init__(
        self,
        db: Database,
        indices: list = None,
        *,
        sort: bool = True,
        name: str | None = None,
    ) -> None:

        self._db = db

        indices = indices or []

        if not isinstance(indices, list):
            indices = list(indices)

        assert all(isinstance(i, int) for i in indices)

        if sort:
            self._indices = sorted(list(set(indices)))
        else:

            # remove duplicates but keep order
            self._indices = dict()
            for i in indices:
                if i not in self._indices:
                    self._indices[i] = i
            self._indices = list(self._indices.keys())

        self._interactions = None

        self._name = name

    ### PROPERTIES

    @property
    def db(self) -> "Database":
        """Returns the associated :class:`.Database`"""
        return self._db

    @property
    def table(self) -> str:
        """Returns the name of the :class:`.Database` table"""
        return self._table

    @property
    def indices(self) -> list[int]:
        """Returns the ids of poses in this set"""
        return self._indices

    @property
    def ids(self) -> list[int]:
        """Returns the ids of poses in this set"""
        return self._indices

    @property
    def name(self) -> str | None:
        """Returns the name of set"""
        return self._name

    @property
    def names(self) -> list[str]:
        """Returns the aliases of poses in this set"""
        return [p.name for p in self]

    @property
    def aliases(self) -> list[str]:
        """Returns the aliases of child poses"""
        return [
            self.db.select_where(
                table=self.table, query="pose_alias", key="id", value=i, multiple=False
            )[0]
            for i in self.indices
        ]

    @property
    def inchikeys(self) -> list[str]:
        """Returns the inchikeys of child poses"""
        return [
            self.db.select_where(
                table=self.table,
                query="pose_inchikey",
                key="id",
                value=i,
                multiple=False,
            )[0]
            for i in self.indices
        ]

    @property
    def id_name_dict(self) -> dict:
        """Return a dictionary mapping pose ID's to their name"""

        records = self.db.select_where(
            table=self.table,
            query="pose_id, pose_inchikey, pose_alias",
            key=f"pose_id IN {self.str_ids}",
            multiple=True,
        )

        lookup = {}
        for i, inchikey, alias in records:
            if alias:
                lookup[i] = alias
            else:
                lookup[i] = inchikey

        return lookup

    @property
    def smiles(self) -> list[str]:
        """Returns the smiles of poses in this set"""
        pairs = self.db.select_where(
            table=self.table,
            query="pose_id, pose_smiles",
            key=f"pose_id IN {self.str_ids}",
            multiple=True,
        )

        results = []
        for pose_id, smiles in pairs:
            if smiles is None:
                pose = self.db.get_pose(id=pose_id)
                smiles = pose.smiles

            results.append(smiles)

        return results

    @property
    def tags(self) -> set[str]:
        """Returns the set of unique tags present in this pose set"""
        values = self.db.select_where(
            table="tag",
            query="DISTINCT tag_name",
            key=f"tag_pose in {self.str_ids}",
            multiple=True,
        )
        return set(v for v, in values)

    @property
    def compounds(self) -> "CompoundSet":
        """Get the compounds associated to this set of poses"""
        from .cset import CompoundSet

        ids = self.db.select_where(
            table="pose",
            query="DISTINCT pose_compound",
            key=f"pose_id in {self.str_ids}",
            multiple=True,
        )
        ids = [v for v, in ids]
        return CompoundSet(self.db, ids)

    @property
    def mols(self) -> "list[rdkit.Chem.mol]":
        """Get the rdkit Molecules contained in this set"""
        return [p.mol for p in self]

    @property
    def num_compounds(self) -> int:
        """Count the compounds associated to this set of poses"""
        return len(self.compounds)

    @property
    def df(self) -> "pandas.DataFrame":
        """Get a DataFrame of the poses in this set"""
        return self.get_df(mol=True)

    @property
    def references(self) -> "PoseSet":
        """Return a :class:`.PoseSet` of the all the distinct references in this :class:`.PoseSet`"""
        return PoseSet(self.db, self.reference_ids)

    @property
    def reference_ids(self) -> set[int]:
        """Return a set of :class:`.Pose` ID's of the all the distinct references in this :class:`.PoseSet`"""
        values = self.db.select_where(
            table="pose",
            query="DISTINCT pose_reference",
            key=f"pose_reference IS NOT NULL and pose_id in {self.str_ids}",
            value=None,
            multiple=True,
        )
        return set(v for v, in values)

    @property
    def inspiration_sets(self) -> list[set[int]]:
        """Return a list of unique sets of inspiration :class:`.Pose` IDs"""

        sql = f"""
        SELECT inspiration_derivative, inspiration_original FROM inspiration
        WHERE inspiration_derivative IN {self.str_ids}
        """

        pairs = self.db.execute(sql).fetchall()

        data = {}
        for derivative, original in pairs:
            if derivative not in data:
                data[derivative] = set()
            data[derivative].add(original)

        data = {k: tuple(sorted(list(v))) for k, v in data.items()}

        unique = set(data.values())

        return unique

    @property
    def num_inspiration_sets(self) -> int:
        """Return the number of unique sets of inspirations"""
        return len(self.inspiration_sets)

    @property
    def num_inspirations(self) -> int:
        """Return the number of unique inspirations for poses in this set"""
        (count,) = self.db.select_where(
            table="inspiration",
            query="COUNT(DISTINCT inspiration_original)",
            key=f"inspiration_derivative IN {self.str_ids}",
        )

        return count

    @property
    def inspirations(self) -> int:
        """Return the number of unique inspirations for poses in this set"""
        records = self.db.select_where(
            table="inspiration",
            query="DISTINCT inspiration_original",
            key=f"inspiration_derivative IN {self.str_ids}",
            multiple=True,
        )

        if not records:
            return None

        return PoseSet(self.db, [i for i, in records])

    @property
    def str_ids(self) -> str:
        """Return an SQL formatted tuple string of the :class:`.Pose` IDs"""
        return str(tuple(self.ids)).replace(",)", ")")

    @property
    def targets(self) -> "list[Target]":
        """Returns the :class:`.Target` objects of poses in this set"""
        return [self.db.get_target(id=q) for q in self.target_ids]

    @property
    def target_names(self) -> list[str]:
        """Returns the :class:`.Target` objects of poses in this set"""
        return [self.db.get_target_name(id=q) for q in self.target_ids]

    @property
    def target_ids(self) -> list[int]:
        """Returns the :class:`.Target` objects ID's of poses in this set"""
        result = self.db.select_where(
            table=self.table,
            query="DISTINCT pose_target",
            key=f"pose_id in {self.str_ids}",
            multiple=True,
        )
        return [q for q, in result]

    @property
    def best_placed_pose(self) -> Pose:
        """Returns the pose with the best distance_score in this subset"""
        return self.db.get_pose(id=self.best_placed_pose_id)

    @property
    def best_placed_pose_id(self) -> int:
        """Get the id of the pose with the best distance_score in this subset"""
        query = f"pose_id, MIN(pose_distance_score)"
        query = self.db.select_where(
            table="pose", query=query, key=f"pose_id in {self.str_ids}", multiple=False
        )
        return query[0]

    @property
    def interactions(self) -> "InteractionSet":
        """Get a :class:`.InteractionSet` for this :class:`.Pose`"""
        if self._interactions is None:
            from .iset import InteractionSet

            self._interactions = InteractionSet.from_pose(self)
        return self._interactions

    @property
    def interaction_overlap_score(self) -> int:
        """Count the number of member pose pairs which share at least one but not all interactions"""

        sql = f"""
        SELECT DISTINCT interaction_pose, feature_id, interaction_type FROM interaction 
        INNER JOIN feature ON interaction_feature = feature_id
        WHERE interaction_pose IN {self.str_ids}
        """

        # mrich.print(sql)

        records = self.db.execute(sql).fetchall()

        ISETS = {}
        for pose_id, feature_id, interaction_type in records:
            values = ISETS.get(pose_id, set())
            values.add((interaction_type, feature_id))
            ISETS[pose_id] = values

        ids = [i for i in self.ids if i in ISETS]

        count = 0
        for pose_j in ids:
            iset_j = ISETS[pose_j]
            for pose_k in ids:
                iset_k = ISETS[pose_k]

                # try:
                # except KeyError:
                #     mrich.error("No interactions for pose with id", pose_k, "Has it been fingerprinted?")
                #     continue

                intersection = iset_j & iset_k
                diff1 = iset_j - iset_k
                diff2 = iset_k - iset_j

                if intersection and diff1 and diff2:
                    count += 1

        return count

    def get_interaction_clusters(self) -> "dict[int, PoseSet]":
        """Cluster poses based on shared interactions."""

        import networkx as nx
        import community as louvain
        from itertools import combinations

        # get interaction records

        sql = f"""
        SELECT DISTINCT interaction_pose, feature_residue_name, feature_residue_number, interaction_type FROM interaction 
        INNER JOIN feature ON interaction_feature = feature_id
        WHERE interaction_pose IN {self.str_ids}
        """

        records = self.db.execute(sql).fetchall()

        ISETS = {}
        for (
            pose_id,
            feature_residue_name,
            feature_residue_number,
            interaction_type,
        ) in records:
            values = ISETS.get(pose_id, set())
            values.add((interaction_type, feature_residue_name, feature_residue_number))
            ISETS[pose_id] = values

        pairs = combinations(ISETS.keys(), 2)

        # construct overlap dictionary

        OVERLAPS = {}
        for id1, id2 in pairs:
            iset1 = ISETS[id1]
            iset2 = ISETS[id2]
            OVERLAPS[(id1, id2)] = len(iset1 & iset2)

        # make the graph
        G = nx.Graph()

        for (id1, id2), count in OVERLAPS.items():
            G.add_edge(id1, id2, weight=count)

        # partition the graph

        partition = louvain.best_partition(G, weight="weight")

        # find the clusters

        clusters = {}
        for node, cluster_id in partition.items():
            clusters.setdefault(cluster_id, set()).add(node)

        # create the PoseSets

        psets = {
            i: PoseSet(self.db, ids, name=f"Cluster {i}")
            for i, ids in enumerate(clusters.values())
        }

        all_ids = set(sum((pset.ids for pset in psets.values()), []))

        # calculate modal interactions

        for i, cluster in psets.items():

            mrich.var(cluster.name, len(cluster), unit="poses")

            df = cluster.interactions.df

            unique_counts = df.groupby(["type", "residue_name", "residue_number"])[
                "pose_id"
            ].nunique()

            max_count = unique_counts.max()
            max_pairs = unique_counts[unique_counts == max_count]

            for (
                interaction_type,
                residue_name,
                residue_number,
            ) in max_pairs.index.values:
                mrich.print(interaction_type, "w/", residue_name, residue_number)

        # unclustered
        unclustered = set((i for i in self.ids if i not in all_ids))
        psets[None] = PoseSet(self.db, unclustered, name="Unclustered")

        return psets

    @property
    def num_fingerprinted(self) -> int:
        """Count the number of fingerprinted poses in this set"""
        return self.db.count_where(
            table="pose", key=f"pose_id IN {self.str_ids} AND pose_fingerprint = 1"
        )

    @property
    def fraction_fingerprinted(self) -> float:
        """Return the fraction of fingerprinted poses in this set"""
        return self.num_fingerprinted / len(self)

    @property
    def num_subsites(self) -> int:
        """Count the number of subsites that poses in this set come into contact with"""
        (count,) = self.db.select_where(
            query="COUNT(DISTINCT subsite_tag_ref)",
            table="subsite_tag",
            key=f"subsite_tag_pose IN {self.str_ids}",
            none="quiet",
        )
        if count is None:
            count = 0
        return count

    @property
    def subsite_balance(self) -> float:
        """Measure of how evenly subsite counts are distributed across poses in this set"""

        from numpy import std

        sql = f"""
        SELECT COUNT(DISTINCT subsite_tag_ref) FROM subsite_tag
        WHERE subsite_tag_pose IN {self.str_ids}
        GROUP BY subsite_tag_pose
        """

        counts = self.db.execute(sql).fetchall()

        counts = [c for c, in counts] + [0 for _ in range(len(self) - len(counts))]

        return -std(counts)

    @property
    def subsite_ids(self) -> set[int]:
        """Return a list of subsite id's of member poses"""

        sql = f"""
        SELECT DISTINCT subsite_tag_ref FROM subsite_tag
        WHERE subsite_tag_pose IN {self.str_ids}
        """

        subsite_ids = self.db.execute(sql).fetchall()

        if not subsite_ids:
            return set()

        subsite_ids = set([i for i, in subsite_ids])

        return subsite_ids

    @property
    def avg_energy_score(self) -> float:
        """Average energy score of poses in this set"""

        from numpy import mean

        sql = f"""
        SELECT pose_energy_score FROM pose
        WHERE pose_id IN {self.str_ids}
        """

        scores = self.db.execute(sql).fetchall()
        return mean([s for s, in scores if s is not None])

    @property
    def avg_distance_score(self) -> float:
        """Average distance score of poses in this set"""

        from numpy import mean

        sql = f"""
        SELECT pose_distance_score FROM pose
        WHERE pose_id IN {self.str_ids}
        """

        scores = self.db.execute(sql).fetchall()

        return mean([s for s, in scores if s is not None])

    @property
    def derivatives(self) -> "PoseSet":
        ids = self.db.select_where(
            table="inspiration",
            query="inspiration_derivative",
            key=f"inspiration_original IN {self.str_ids}",
            multiple=True,
            none="quiet",
        )
        if not ids:
            return None
        ids = [i for i, in ids]
        pset = PoseSet(self.db, ids, name=f"derivatives of {self}")
        return pset

    ### FILTERING

    def get_by_tag(
        self,
        tag: str,
        inverse: bool = False,
    ) -> "PoseSet":
        """Get all child poses with a certain tag

        :param tag: tag to filter by
        :param inverse: return all poses *not* tagged with ``tag`` (Default value = False)

        """
        values = self.db.select_where(
            query="tag_pose", table="tag", key="name", value=tag, multiple=True
        )
        if inverse:
            matches = [v for v, in values if v]
            ids = [i for i in self.ids if i not in matches]
        else:
            ids = [v for v, in values if v and v in self.ids]
        return PoseSet(self.db, ids)

    def get_by_metadata(
        self, key: str, value: str | None = None, debug: bool = False
    ) -> "PoseSet":
        """Get all child poses with by their metadata. If no value is passed, then simply containing the key in the metadata dictionary is sufficient

        :param key: metadata key to search for
        :param value: metadata value, if ``None`` return poses with the metadata key regardless of value (Default value = None)

        """
        results = self.db.select_where(
            query="pose_id, pose_metadata",
            key=f"pose_id IN {self.str_ids}",
            table="pose",
            multiple=True,
        )

        if value is None:
            ids = [i for i, d in results if d and f'"{key}":' in d]

        else:
            if isinstance(value, str):
                value = f'"{value}"'

            ids = []

            for i, d in results:
                if not d:
                    continue

                if debug:
                    mrich.print(i, d, f'"{key}": {value}' in d)

                if f'"{key}": {value}' in d:
                    ids.append(i)
                else:
                    continue

                if debug:
                    break

        return PoseSet(self.db, ids)

    def get_by_inspiration(self, inspiration: int | Pose, inverse: bool = False):
        """Get all child poses with with this inspiration.

        :param inspiration: inspiration :class:`.Pose` ID or object
        :param inverse: invert the selection (Default value = False)

        """

        ids = set()

        for pose in self:
            if not inverse:
                for pose_inspiration in pose.inspirations:
                    if pose_inspiration == inspiration:
                        ids.add(pose.id)
                        break

            elif inverse:
                for pose_inspiration in pose.inspirations:
                    if pose_inspiration == inspiration:
                        break
                else:
                    ids.add(pose.id)

        return PoseSet(self.db, ids)

    def get_df(
        self,
        smiles: bool = True,
        inchikey: bool = True,
        alias: bool = True,
        name: bool = True,
        compound_id: bool = False,
        target_id: bool = False,
        reference_id: bool = False,
        path: bool = False,
        mol: bool = False,
        energy_score: bool = False,
        distance_score: bool = False,
        inspiration_score: bool = False,
        metadata: bool = False,
        expand_metadata: bool = True,
        debug: bool = True,
        inspiration_ids: bool = False,
        inspiration_aliases: bool = False,
        derivative_ids: bool = False,
        tags: bool = False,
        subsites: bool = False,
        # skip_no_mol=True, reference: str = "name", mol: bool = False, **kwargs
    ) -> "pandas.DataFrame":
        """Get a DataFrame of the poses in this set.

        :param smiles: include SMILES column (Default value = True)
        :param inchikey: include InChIKey column (Default value = True)
        :param alias: include alias column (Default value = True)
        :param name: include name column (Default value = True)
        :param compound_id: include :class:`.Compound` ID column (Default value = False)
        :param reference_id: include reference :class:`.Pose` ID column (Default value = False)
        :param target_id: include reference :class:`.Target` ID column (Default value = False)
        :param path: include path column (Default value = False)
        :param mol: include ``rdkit.Chem.Mol`` in output (Default value = False)
        :param energy_score: include energy_score column (Default value = False)
        :param distance_score: include distance_score column (Default value = False)
        :param inspiration_score: include inspiration_score column (Default value = False)
        :param metadata: include metadata in output (Default value = False)
        :param expand_metadata: create separate column for each metadata key (Default value = True)
        :param inspiration_ids: include inspiration :class:`.Pose` ID column
        :param inspiration_aliases: include inspiration :class:`.Pose` alias column
        :param derivative_ids: include derivative :class:`.Pose` ID column
        :param tags: include tags column
        :param subsites: include subsites column
        """

        from json import loads
        from rdkit.Chem import Mol
        from pandas import DataFrame

        get_alias = alias

        if name:
            alias = True

        query = ["pose_id"]

        if smiles:
            query.append("pose_smiles")

        if inchikey:
            query.append("pose_inchikey")

        if alias:
            query.append("pose_alias")

        if reference_id:
            query.append("pose_reference")

        if path:
            query.append("pose_path")

        if compound_id:
            query.append("pose_compound")

        if target_id:
            query.append("pose_target")

        if mol:
            query.append("pose_mol")

        if energy_score:
            query.append("pose_energy_score")

        if distance_score:
            query.append("pose_distance_score")

        if inspiration_score:
            query.append("pose_inspiration_score")

        if metadata:
            query.append("pose_metadata")

        query = ", ".join(query)

        sql = f"""
        SELECT {query}
        FROM pose
        WHERE pose_id IN {self.str_ids}
        """

        if debug:
            # print(sql)
            mrich.debug("querying...")
        records = self.db.execute(sql).fetchall()

        if debug:
            generator = mrich.track(records)
        else:
            generator = records

        data = []
        for row in generator:

            row = list(row)

            d = dict(id=row.pop(0))

            if smiles:
                d["smiles"] = row.pop(0)

            if inchikey:
                d["inchikey"] = row.pop(0)

            if alias:
                d["alias"] = row.pop(0)

            if reference_id:
                d["reference_id"] = row.pop(0)

            if path:
                d["path"] = row.pop(0)

            if compound_id:
                d["compound_id"] = row.pop(0)

            if target_id:
                d["target_id"] = row.pop(0)

            if mol:
                d["mol"] = Mol(row.pop(0))

            if energy_score:
                d["energy_score"] = row.pop(0)

            if distance_score:
                d["distance_score"] = row.pop(0)

            if inspiration_score:
                d["inspiration_score"] = row.pop(0)

            if metadata and (meta_str := row.pop(0)):

                meta_dict = loads(meta_str) or {}

                if expand_metadata:
                    for k, v in meta_dict.items():
                        d[k] = v

                else:
                    d["metadata"] = meta_dict

            data.append(d)

        df = DataFrame(data)

        if inspiration_ids or derivative_ids or inspiration_aliases:
            if debug:
                mrich.debug("adding inspiration column(s)")

            tuples = self.db.get_inspiration_tuples()

            if inspiration_ids or inspiration_aliases:
                lookup = {}
                for inspiration, derivative in tuples:
                    lookup.setdefault(derivative, set())
                    lookup[derivative].add(inspiration)
                df["inspiration_ids"] = df["id"].apply(lambda x: lookup.get(x, {}))

            if derivative_ids:
                lookup = {}
                for inspiration, derivative in tuples:
                    lookup.setdefault(inspiration, set())
                    lookup[inspiration].add(derivative)
                df["derivative_ids"] = df["id"].apply(lambda x: lookup.get(x, {}))

        if inspiration_aliases:
            inspirations = PoseSet(
                self.db, set.union(*list(df["inspiration_ids"].values))
            )
            lookup = self.db.get_pose_id_alias_dict(pset=inspirations)
            inspiration_aliases = []
            for ids in df["inspiration_ids"].values:
                aliases = {lookup[i] for i in ids}
            df["inspiration_aliases"] = df["inspiration_ids"].apply(
                lambda x: {lookup[i] for i in x}
            )
            if not inspiration_ids:
                df = df.drop(columns=["inspiration_ids"])

        if tags:
            if debug:
                mrich.debug("adding tag column")
            lookup = self.db.get_pose_tag_dict()
            df["tags"] = df["id"].apply(lambda x: lookup.get(x, {}))

        if subsites:
            if debug:
                mrich.debug("adding subsite column")
            lookup = self.db.get_pose_subsite_names_dict()
            df["subsites"] = df["id"].apply(lambda x: lookup.get(x, set()))

        if name:
            df["name"] = df.apply(lambda row: row["alias"] or f'P{row["id"]}', axis=1)
            if not get_alias:
                df = df.drop(columns=["alias"])

        df = df.set_index("id")

        return df

    def get_by_reference(
        self,
        ref_id: int,
    ) -> "PoseSet | None":
        """Get poses with a certain reference id

        :param ref_id: reference :class:`.Pose` ID

        """
        values = self.db.select_where(
            table="pose",
            query="pose_id",
            key=f"pose_reference={ref_id} AND pose_id in {self.str_ids}",
            multiple=True,
        )
        if not values:
            return None
        return PoseSet(self.db, [v for v, in values])

    def get_by_compound(
        self,
        *,
        compound: "int | Compound",
    ) -> "PoseSet | None":
        """Select a subset of this :class:`.PoseSet` by the associated :class:`.Compound`.

        :param compound: :class:`.Compound` object or ID
        :returns: a :class:`.PoseSet` of the selection

        """
        from .compound import Compound

        if isinstance(compound, Compound):
            compound = compound.id

        values = self.db.select_where(
            query="pose_id",
            table="pose",
            key=f"pose_compound={compound} AND pose_id in {self.str_ids}",
            multiple=True,
            none="quiet",
        )
        if not values:
            return None
        ids = [v for v, in values if v]
        return PoseSet(self.db, [v for v, in values])

    def get_by_target(
        self,
        *,
        id: int,
    ) -> "PoseSet | None":
        """Select a subset of this :class:`.PoseSet` by the associated :class:`.Target`.

        :param id: :class:`.Target` ID
        :returns: a :class:`.PoseSet` of the selection

        """
        assert isinstance(id, int)
        values = self.db.select_where(
            query="pose_id",
            table="pose",
            key=f"pose_target is {id} AND pose_id in {self.str_ids}",
            multiple=True,
            none="quiet",
        )
        ids = [v for v, in values if v]
        if not ids:
            return None
        return PoseSet(self.db, ids)

    def get_by_subsite(
        self,
        *,
        id: int,
    ) -> "PoseSet | None":
        """Select a subset of this :class:`.PoseSet` by the associated :class:`.Subsite`.

        :param id: :class:`.Subsite` ID
        :returns: a :class:`.PoseSet` of the selection

        """
        assert isinstance(id, int)
        values = self.db.select_where(
            query="subsite_tag_pose",
            table="subsite_tag",
            key=f"subsite_tag_ref is {id} AND subsite_tag_pose in {self.str_ids}",
            multiple=True,
            none="quiet",
        )
        ids = [v for v, in values if v]
        if not ids:
            return None

        if self.name:
            name = f"{self.name} & subsite={id}"
        else:
            name = None

        return PoseSet(self.db, ids, name=name)

    def get_best_placed_poses_per_compound(self):
        """Choose the best placed pose (best distance_score) grouped by compound"""

        sql = f"""
        SELECT pose_id, MIN(pose_distance_score) FROM pose
        WHERE pose_id IN {self.str_ids}
        GROUP BY pose_compound
        """

        cursor = self.db.execute(sql)

        ids = [i for i, _ in cursor]

        return PoseSet(self.db, ids)

    def filter(
        self,
        function=None,
        *,
        key: str = None,
        value: str = None,
        operator="=",
        inverse: bool = False,
    ):
        """Filter this :class:`.PoseSet` by selecting members where ``function(pose)`` is truthy or pass a key, value, and optional operator to search by database values

        :param function: callable object
        :param key: database field for 'pose' table ('pose_' prefix not needed)
        :param value: value to compare to
        :param operator: comparison operator (default = "=")
        :param inverse: invert the selection (Default value = False)

        """

        if function:

            ids = set()
            for pose in self:
                value = function(pose)
                # mrich.debug(f'{pose=} {value=}')
                if value and not inverse:
                    ids.add(pose.id)
                elif not value and inverse:
                    ids.add(pose.id)

            return PoseSet(self.db, ids)

        sql = f"""
        SELECT pose_id FROM pose
        WHERE pose_id IN {self.str_ids}
        AND pose_{key} {operator} {value}
        """

        cursor = self.db.execute(sql)

        ids = [i for i, in cursor]

        return PoseSet(self.db, ids)

    ### BULK SETTING

    @property
    def reference(self):
        """Bulk set the references for poses in this set"""
        raise NotImplementedError(
            "This attribute only allows setting, ``PoseSet.reference = ...``"
        )

    @reference.setter
    def reference(self, r) -> None:
        """Bulk set the references for poses in this set"""
        if not isinstance(r, int):
            assert r._table == "pose"
            r = r.id

        for i in self.indices:
            self.db.update(
                table="pose", id=i, key="pose_reference", value=r, commit=False
            )

        self.db.commit()

    def add_tag(
        self,
        tag: str,
    ) -> None:
        """Add this tag to every member of the set"""

        assert isinstance(tag, str)

        for i in self.indices:
            self.db.insert_tag(name=tag, pose=i, commit=False)

        mrich.print(f'Tagged {self} w/ "{tag}"')

        self.db.commit()

    def append_to_metadata(
        self,
        key,
        value,
    ) -> None:
        """Append a specific item to list-like values associated with a given key for all member's metadata dictionaries

        :param key: the :class:`.Metadata` key to match
        :param value: the value to append to the list

        """
        for id in self.indices:
            metadata = self.db.get_metadata(table="pose", id=id)
            metadata.append(key, value)

    def set_subsites_from_metadata_field(self, field="CanonSites alias") -> None:
        """Create and assign subsite entries from a metadata field

        :param field: the metadata field to use

        """

        from json import loads

        records = self.db.select_where(
            table="pose",
            query="pose_id, pose_target, pose_metadata",
            key=f"pose_id IN {self.str_ids}",
            multiple=True,
        )

        subsites = set()
        subsite_tags = set()

        for pose_id, pose_target, metadata in records:

            metadata = loads(metadata)

            key = metadata.get(field)

            if not key:
                mrich.warning(field, "not in metadata pose_id=", pose_id)
                continue

            subsites.add((pose_target, key))
            subsite_tags.add((key, pose_id))

        sql = """
        INSERT OR IGNORE INTO subsite(subsite_target, subsite_name)
        VALUES(?1, ?2)
        RETURNING subsite_id
        """

        records = self.db.executemany(sql, sorted(list(subsites)))
        subsite_ids = [i for i, in records]
        subsite_lookup = {name: i for (t, name), i in zip(subsites, subsite_ids)}

        # supplement existing subsites
        subsite_lookup.update(
            {
                n: i
                for i, n in self.db.select(
                    table="subsite", query="subsite_id, subsite_name", multiple=True
                )
            }
        )

        sql = """
        INSERT OR IGNORE INTO subsite_tag(subsite_tag_ref, subsite_tag_pose)
        VALUES(?1, ?2)
        """

        subsite_tags = [
            (subsite_lookup[subsite], pose_id) for subsite, pose_id in subsite_tags
        ]

        self.db.executemany(sql, subsite_tags)

        self.db.commit()

    def calculate_inspiration_scores(
        self,
        alpha: float = 0.95,
        beta: float = 0.05,
        score_type: str = "combo",
    ) -> "pd.DataFrame":
        """Set inspiration_score values using MoCASSIn.calculate_mocassin_tversky

        :param alpha: Tversky alpha parameter
        :param beta: Tversky beta parameter
        :param score_type: Score type to add to database, choose from "combo", "shape", "colour"
        :returns: Pandas DataFrame with molecules and scores
        """

        from mocassin.mocassin import calculate_mocassin_tversky

        df = self.get_df(
            alias=False,
            smiles=False,
            inchikey=False,
            inspiration_ids=True,
            mol=True,
        )

        inspirations = {p.id: p for p in self.inspirations}

        df["inspiration_mols"] = df["inspiration_ids"].apply(
            lambda x: [inspirations[i].mol for i in x]
        )

        n = len(df)

        for j, (i, row) in mrich.track(
            enumerate(df.iterrows()), prefix="MoCASSIn", total=n
        ):

            mrich.set_progress_field("j", j)
            mrich.set_progress_field("n", n)

            try:
                combo, shape, colour = calculate_mocassin_tversky(
                    row["inspiration_mols"],
                    row["mol"],
                    alpha=0.95,
                    beta=0.05,
                )
                df.loc[i, f"mocassin_combo({alpha},{beta})"] = combo
                df.loc[i, f"mocassin_shape({alpha},{beta})"] = shape
                df.loc[i, f"mocassin_colour({alpha},{beta})"] = colour
            except Exception as e:
                mrich.error(e)

        tuples = df[f"mocassin_{score_type}({alpha},{beta})"].items()

        sql = """UPDATE pose SET pose_inspiration_score = ?2 WHERE pose_id = ?1"""

        mrich.debug("Updating pose_inspiration_score values")
        self.db.executemany(sql, tuples)
        self.db.commit()

        return df

    ### SPLITTING

    def split_by_reference(self) -> "dict[int,PoseSet]":
        """Split this :class:`.PoseSet` into subsets grouped by reference ID

        :returns: a dictionary with reference :class:`.Pose` IDs as keys and :class:`.PoseSet` subsets as values

        """
        sets = {}
        for ref_id in self.reference_ids:
            sets[ref_id] = self.get_by_reference(ref_id)
        return sets

    def split_by_inspirations(
        self,
        single_set: bool = False,
    ) -> "dict[int,PoseSet] | PoseSet":
        """Split this :class:`.PoseSet` into subsets grouped by inspirations

        :param single_set: Return a single :class:`.PoseSet` with members sorted by inspirations (Default value = False)
        :returns: a dictionary with tuples of inspiration :class:`.Pose` IDs as keys and :class:`.PoseSet` subsets as values

        """

        sets = {}

        lookup = self.db.get_pose_id_inspiration_ids_dict(pset=self)

        for pose_id, insp_ids in lookup.items():
            key = tuple(insp_ids)
            sets.setdefault(key, set())
            sets[key].add(pose_id)

        mrich.var("#unique inspiration combinations", len(sets))

        if single_set:
            return PoseSet(self.db, sum([s.ids for s in sets.values()], []), sort=False)

        return {
            PoseSet(self.db, insp_ids): PoseSet(self.db, pose_ids)
            for insp_ids, pose_ids in sets.items()
        }

    ### EXPORTING

    def write_sdf(
        self,
        out_path: str,
        name_col: str = "alias",
        inspiration_ids: bool = False,
        inspiration_aliases: bool = False,
        **kwargs,
    ) -> None:
        """Write an SDF

        :param out_path: filepath of the output
        :param name_col: pose property to use as the name column, can be ``["name", "alias", "inchikey", "id"]`` (Default value = 'name')
        :param inspiration_ids: include inspiration :class:`.Pose` ID column
        :param inspiration_aliases: include inspiration :class:`.Pose` alias column
        :param fragalysis_inspirations: create inspirations column "ref_mols"
        """

        from pathlib import Path
        import json

        df = self.get_df(
            mol=True,
            inspiration_ids=inspiration_ids,
            inspiration_aliases=inspiration_aliases,
            name=name_col == "name",
            **kwargs,
        )

        if name_col not in ["name", "alias", "inchikey", "id"]:
            # try getting name from metadata
            records = self.db.select_where(
                table="pose",
                query="pose_id, pose_metadata",
                key=f"pose_id IN {self.str_ids}",
                multiple=True,
            )

            longcode_lookup = {}
            for i, d in records:
                if d:
                    metadata = json.loads(d)
                else:
                    metadata = {}

                longcode_lookup[i] = metadata.get(name_col, None)

            values = []
            for i, row in df.iterrows():
                values.append(longcode_lookup[row["id"]])

            df[name_col] = values

        df.rename(inplace=True, columns={name_col: "_Name", "mol": "ROMol"})

        mrich.writing(out_path)

        from rdkit.Chem import PandasTools

        PandasTools.WriteSDF(df, out_path, "ROMol", "_Name", list(df.columns))

        # keep record of export
        value = str(Path(out_path).resolve())
        self.db.remove_metadata_list_item(table="pose", key="exports", value=value)
        self.append_to_metadata(key="exports", value=value)

    def to_fragalysis(
        self,
        out_path: str,
        *,
        method: str,
        ref_url: str = "https://hippo.winokan.com",
        submitter_name: str,
        submitter_email: str,
        submitter_institution: str,
        metadata: bool = True,
        sort_by: str | None = None,
        sort_reverse: bool = False,
        generate_pdbs: bool = False,
        copy_reference_pdbs: bool = False,
        # ingredients: IngredientSet = None,
        skip_no_reference: bool = True,
        skip_no_inspirations: bool = True,
        skip_metadata: list[str] | None = None,
        tags: bool = True,
        subsites: bool = True,
        extra_cols: dict[str, list] = None,
        # name_col: str = "name",
        **kwargs,
    ):
        """Prepare an SDF for upload to the RHS of Fragalysis.

        :param out_path: the file path to write to
        :param method: method used to generate the compounds
        :param ref_url: reference URL for the method
        :param submitter_name: name of the person submitting the compounds
        :param submitter_email: email of the person submitting the compounds
        :param submitter_institution: institution name of the person submitting the compounds
        :param metadata: include metadata in the output? (Default value = True)
        :param skipmetadata: exclude metadata keys from output
        :param sort_by: if set will sort the SDF by this column/field (Default value = None)
        :param sort_reverse: reverse the sorting (Default value = False)
        :param generate_pdbs: generate accompanying protein-ligand complex PDBs (Default value = False)
        :param ingredients: get procurement and amount information from this :class:`.IngredientSet` (Default value = None)
        :param tags: include a column for tags in the output (Default value = True)
        :param subsites: include a column for subsites in the output (Default value = True)
        :param extra_cols: extra_cols should be a dictionary with a key for each column name, and list values where the first element is the field description, and all subsequent elements are values for each pose.
        :param name: How to determine the molecule name, see :meth:`.PoseSet.get_df`

        """

        from .fragalysis import generate_header
        from pathlib import Path
        from rdkit.Chem import SDWriter, PandasTools

        assert out_path.endswith(".sdf")

        _name_col = "_Name"
        mol_col = "ROMol"

        # make sure references are defined:

        mrich.debug(len(self), "poses in set")
        poses = None

        if skip_no_reference:

            values = self.db.select_where(
                table="pose",
                query="DISTINCT pose_id",
                key=f"pose_reference IS NOT NULL and pose_id in {self.str_ids}",
                multiple=True,
                none="error",
            )

            if not values:
                return

            poses = PoseSet(self.db, [i for i, in values])

            mrich.debug(len(poses), "remaining after skipping null reference")

        if skip_no_inspirations:

            if not poses:
                poses = self

            values = self.db.select_where(
                table="inspiration",
                query="DISTINCT inspiration_derivative",
                key=f"inspiration_derivative IN {poses.str_ids}",
                multiple=True,
                none="error",
            )

            if not values:
                return

            poses = PoseSet(self.db, [i for i, in values])

            mrich.debug(len(poses), "remaining after skipping null inspirations")

        if not poses:
            poses = PoseSet(self.db, self.ids)

        mrich.var("#poses", len(poses))

        # get the dataframe of poses

        pose_df = poses.get_df(
            mol=True,
            inspiration_ids=True,
            # duplicate_name="original ID",
            name=True,
            compound_id=True,
            reference_id=True,
            metadata=metadata,
            tags=tags,
            subsites=subsites,
            energy_score=True,
            distance_score=True,
            inspiration_score=True,
            # sanitise_null_metadata_values=True,
            # sanitise_tag_list_separator=";",
            # sanitise_metadata_list_separator=";",
            # skip_metadata=skip_metadata,
            # **kwargs,
        )

        pose_df = pose_df.reset_index()

        # fix inspirations and reference column (comma separated aliases)

        lookup = self.db.get_pose_id_alias_dict()

        inspiration_strs = []
        for i, row in pose_df.iterrows():
            strs = []
            for i in row["inspiration_ids"]:
                alias = lookup.get(i)
                if not alias:
                    continue
                strs.append(alias)
            inspiration_strs.append(",".join(strs))

        # comma separate subsites
        if subsites:
            pose_df["subsites"] = pose_df["subsites"].apply(lambda x: ",".join(x))

        if tags:
            pose_df["tags"] = pose_df["tags"].apply(lambda x: ",".join(x))

        pose_df["ref_mols"] = inspiration_strs
        pose_df["ref_pdb"] = pose_df["reference_id"].apply(lambda x: lookup[x])

        # add compound identifier column (inchikey?)

        drops = ["inspiration_ids", "reference_id"]

        # if ingredients:
        #     drops.pop(drops.index("compound"))

        if skip_no_reference:
            prev = len(pose_df)
            pose_df = pose_df[pose_df["reference_id"].notna()]
            if len(pose_df) < prev:
                mrich.warning(f"Skipping {prev - len(pose_df)} Poses with no reference")

        pose_df = pose_df.drop(columns=drops, errors="ignore")

        pose_df[_name_col] = pose_df["name"]

        pose_df.rename(
            inplace=True,
            columns={
                "id": "HIPPO Pose ID",
                "compound_id": "HIPPO Compound ID",
                "mol": mol_col,
                # "smiles": "original SMILES",
                # "compound_id": "compound inchikey",
            },
        )

        extras = {
            "HIPPO Pose ID": "HIPPO Pose ID",
            "HIPPO Compound ID": "HIPPO Compound ID",
            "smiles": "smiles",
            "ref_pdb": "protein reference",
            "ref_mols": "fragment inspirations",
            "alias": "alias",
            # "compound inchikey": "compound inchikey",
            "distance_score": "distance_score",
            "energy_score": "energy_score",
            "inspiration_score": "inspiration_score",
        }

        if subsites:
            extras["subsites"] = "subsites"

        if tags:
            extras["tags"] = "tags"

        if extra_cols:
            for key, value in extra_cols.items():
                extras[key] = value[0]

        # if ingredients:

        #     q_entries = []
        #     q_prices = []
        #     q_lead_times = []
        #     q_amounts = []

        #     currency = None

        #     for i, row in pose_df.iterrows():

        #         compound_id = self.db.get_compound_id(inchikey=row["compound inchikey"])

        #         ingredient = ingredients(compound_id=compound_id)

        #         if isinstance(ingredient, IngredientSet):
        #             ingredient = sorted(
        #                 [i for i in ingredient], key=lambda x: x.quote.price
        #             )[0]

        #         quote = ingredient.quote
        #         if not currency:
        #             currency = quote.currency
        #         else:
        #             assert quote.currency == currency

        #         q_entries.append(quote.entry_str)
        #         q_prices.append(quote.price)
        #         q_lead_times.append(quote.lead_time)
        #         q_amounts.append(quote.amount)

        #     pose_df["Supplier Catalogue Entry"] = q_entries
        #     # pose_df['Supplier:Catalogue:Entry'] = q_entries
        #     pose_df[f"Price ({currency})"] = q_prices
        #     pose_df["Lead time (working days)"] = q_lead_times
        #     pose_df["Amount (mg)"] = q_amounts

        #     extras["Supplier Catalogue Entry"] = "Supplier Catalogue Entry string"
        #     extras[f"Price ({currency})"] = "Quoted price"
        #     extras["Lead time (working days)"] = "Quoted lead-time"
        #     extras["Amount (mg)"] = "Quoted amount"

        out_path = Path(out_path).resolve()
        mrich.var("out_path", out_path)

        if generate_pdbs:

            from zipfile import ZipFile

            # output subdirectory
            out_key = Path(out_path).name.removesuffix(".sdf")
            pdb_dir = Path(out_path).parent / Path(out_key)
            pdb_dir.mkdir(exist_ok=True)
            zip_path = Path(out_path).parent / f"{out_key}_pdbs.zip"

            # create the zip archive
            with ZipFile(str(zip_path.resolve()), "w") as z:

                # loop over poses
                for (i, row), pose in zip(pose_df.iterrows(), poses):

                    # filenames
                    pdb_name = f"{out_key}_{row._Name}.pdb"
                    pdb_path = pdb_dir / pdb_name
                    pose_df.loc[i, "ref_pdb"] = pdb_name

                    # generate the PL-complex
                    sys = pose.complex_system

                    # write the PDB
                    mrich.writing(pdb_path)
                    sys.write(pdb_path, verbosity=0)
                    z.write(pdb_path)

            mrich.writing(f"{out_key}_pdbs.zip")

        if copy_reference_pdbs:

            from zipfile import ZipFile
            import shutil

            # output subdirectory
            out_key = Path(out_path).name.removesuffix(".sdf")
            pdb_dir = Path(out_path).parent / Path(out_key)
            pdb_dir.mkdir(exist_ok=True)
            zip_path = Path(out_path).parent / f"{out_key}_refs.zip"

            references = self.references
            lookup = self.db.get_pose_alias_path_dict(references)

            zips = set()
            for ref_alias in pose_df["ref_pdb"].values:
                source_path = Path(lookup[ref_alias])

                apo_path = source_path.parent / source_path.name.replace(
                    "_hippo.pdb", ".pdb"
                ).replace(".pdb", "_apo-desolv.pdb")

                if not apo_path.exists():
                    sys = mp.parse(source_path).protein_system
                    sys.write(apo_path, verbosity=0)

                target_path = pdb_dir / f"{ref_alias}.pdb"

                if not target_path.exists():
                    mrich.writing(target_path)
                    shutil.copy(apo_path, target_path)

                zips.add(target_path)

            # create the zip archive
            with ZipFile(str(zip_path.resolve()), "w") as z:
                for path in zips:
                    z.write(path, arcname=path.name)

            mrich.writing(f"{out_key}_refs.zip")

        # create the header molecule

        df_cols = set(pose_df.columns)

        header = generate_header(
            self[0],
            method=method,
            ref_url=ref_url,
            submitter_name=submitter_name,
            submitter_email=submitter_email,
            submitter_institution=submitter_institution,
            extras=extras,
            metadata=metadata,
        )

        header_cols = set(header.GetPropNames())

        # # empty properties
        # pose_df["generation_date"] = [None] * len(pose_df)
        # pose_df["submitter_name"] = [None] * len(pose_df)
        # pose_df["method"] = [None] * len(pose_df)
        # pose_df["submitter_email"] = [None] * len(pose_df)
        # pose_df["ref_url"] = [None] * len(pose_df)

        if extra_cols:
            for key, value in extra_cols.items():
                if len(value) != len(pose_df) + 1:
                    mrich.error(
                        f'extra_col "{key}" does not have the correct number of values'
                    )
                    raise ValueError(
                        f'extra_col "{key}" does not have the correct number of values'
                    )
                pose_df[key] = value[1:]

        if sort_by:
            pose_df = pose_df.sort_values(by=sort_by, ascending=not sort_reverse)

        fields = []

        mrich.writing(out_path)

        with open(out_path, "w") as sdfh:
            with SDWriter(sdfh) as w:
                w.write(header)
            PandasTools.WriteSDF(
                pose_df, sdfh, mol_col, _name_col, set(pose_df.columns)
            )

        # keep record of export
        value = str(Path(out_path).resolve())
        self.db.remove_metadata_list_item(table="pose", key="exports", value=value)
        self.append_to_metadata(key="exports", value=value)

        return pose_df

    def to_pymol(self, prefix: str | None = None) -> None:
        """Group the poses by reference protein and inspirations and output relevant PDBs and SDFs.

        :param prefix: prefix to give all output subdirectories (Default value = None)

        """

        commands = []

        prefix = prefix or ""
        if prefix:
            prefix = f"{prefix}_"

        from pathlib import Path

        for i, (ref_id, poses) in enumerate(self.split_by_reference().items()):

            ref_pose = self.db.get_pose(id=ref_id)
            ref_name = ref_pose.name or ref_id

            # create the subdirectory
            ref_dir = Path(f"{prefix}ref_{ref_name}")
            mrich.writing(ref_dir)
            ref_dir.mkdir(parents=True, exist_ok=True)

            # write the reference protein
            ref_pdb = ref_dir / f"ref_{ref_name}.pdb"
            ref_pose.protein_system.write(ref_pdb, verbosity=0)

            # color the reference:
            commands.append(f"load {ref_pdb.resolve()}")
            commands.append("hide")
            commands.append("show lines")
            commands.append("show surface")
            commands.append("util.cbaw")
            commands.append("set surface_color, white")
            commands.append("set transparency,  0.4")

            for j, (insp_ids, poses) in enumerate(
                poses.split_by_inspirations().items()
            ):

                inspirations = PoseSet(self.db, insp_ids)
                insp_names = "-".join(inspirations.names)

                # create the subdirectory
                insp_dir = ref_dir / insp_names
                insp_dir.mkdir(parents=True, exist_ok=True)

                # write the inspirations
                insp_sdf = insp_dir / f"{insp_names}_frags.sdf"
                inspirations.write_sdf(insp_sdf)

                commands.append(f"load {insp_sdf.resolve()}")
                commands.append(
                    f"set all_states, on, {insp_sdf.name.removesuffix('.sdf')}"
                )
                commands.append(
                    f"util.rainbow \"{insp_sdf.name.removesuffix('.sdf')}\""
                )

                # write the poses
                pose_sdf = insp_dir / f"{insp_names}_derivatives.sdf"
                poses.write_sdf(pose_sdf)

                commands.append(f"load {pose_sdf.resolve()}")
                commands.append(f'util.cbaw "{pose_sdf.name.removesuffix(".sdf")}"')

                if j > 0:
                    commands.append(f"disable \"{insp_sdf.name.removesuffix('.sdf')}\"")
                    commands.append(f'disable "{pose_sdf.name.removesuffix(".sdf")}"')

        return "; ".join(commands)

    def to_knitwork(
        self, out_path: str, path_root: str = ".", aligned_files_dir: str | None = None
    ) -> None:
        """Knitwork takes a CSV input with:

        - observation shortcode
        - smiles
        - path_to_ligand_mol
        - path_to_pdb

        :param out_path: path to output CSV
        :param path_root: paths in CSV will be relative to here

        """

        from os.path import relpath
        from pathlib import Path

        out_path = Path(out_path).resolve()
        path_root = Path(path_root).resolve()
        mrich.var("out_path", out_path)
        mrich.var("path_root", path_root)
        mrich.var("aligned_files_dir", aligned_files_dir)

        assert out_path.name.endswith(".csv")

        with open(out_path, "wt") as f:

            mrich.writing(out_path)

            for pose in self:

                assert pose.alias
                assert "hits" in pose.tags

                if aligned_files_dir:

                    mol = str(pose.mol_path)
                    pdb = str(pose.apo_path)

                    assert "aligned_files" in mol
                    assert "aligned_files" in pdb

                    mol = mol.split("aligned_files/")[-1]
                    pdb = pdb.split("aligned_files/")[-1]

                    aligned_files_dir = Path(aligned_files_dir)

                    mol = relpath(aligned_files_dir / mol, path_root)
                    pdb = relpath(aligned_files_dir / pdb, path_root)

                else:
                    mol = relpath(pose.mol_path, path_root)
                    pdb = relpath(pose.apo_path, path_root)

                data = [pose.alias, pose.compound.smiles, mol, pdb]

                f.write(",".join(data))
                f.write("\n")

    def to_syndirella(
        self, out_key: "str | Path", separate: bool = False
    ) -> "DataFrame":
        """Create syndirella inputs"""

        from pathlib import Path

        out_key = Path(".") / out_key

        if separate:
            dfs = []
            from pandas import concat

            for i, pose in enumerate(self):
                mrich.h3(f"{i}/{len(self)}: {pose}")
                this_out_key = out_key / f"P{pose.id}"
                df = pose.to_syndirella(out_key=this_out_key)
                dfs.append(df)
            return concat(dfs)

        import shutil
        from pandas import DataFrame

        out_dir = out_key.parent
        out_key = out_key.name

        out_dir.mkdir(parents=True, exist_ok=True)

        template_dir = out_dir / "templates"
        mrich.writing(template_dir)
        template_dir.mkdir(parents=True, exist_ok=True)

        # mrich.var("out_dir", out_dir)
        mrich.var("out_key", out_key)
        mrich.var("#poses", len(self))

        pset_inspirations = set()

        data = []
        for pose in mrich.track(self, prefix="preparing inputs..."):

            comp = pose.compound

            ref = pose.reference
            inspirations = pose.inspirations

            if not ref:
                mrich.warning(pose, "has no reference, using self as template")
                ref = pose
                assert ref.apo_path, f"Reference {ref} has no apo_path"

            if not inspirations:
                mrich.warning(pose, "has no inspirations, using self")
                inspirations = PoseSet(self.db, [pose.id])

            for i in inspirations.ids:
                pset_inspirations.add(i)

            d = dict(
                smiles=comp.smiles,
                template=ref.name,
                compound_set=out_key,
            )

            template = template_dir / ref.apo_path.name
            if not template.exists():
                mrich.writing(template)
                shutil.copy(ref.apo_path, template)

            for i, p in enumerate(inspirations):
                d[f"hit{i+1}"] = p.name

            data.append(d)

        df = DataFrame(data)

        csv_name = out_dir / f"{out_key}_syndirella_input.csv"
        mrich.writing(csv_name)
        df.to_csv(csv_name, index=False)

        inspirations = PoseSet(self.db, pset_inspirations)

        sdf_name = out_dir / f"{out_key}_syndirella_inspiration_hits.sdf"
        inspirations.write_sdf(
            sdf_name,
            tags=False,
            metadata=False,
            name_col="name",
        )

        return df

    ### OUTPUT

    def interactive(
        self,
        print_name: str = True,
        method: str | None = None,
        function: Callable | None = None,
        **kwargs,
    ):
        """Interactive widget to navigate compounds in the table

        :param print_name: print the :class:`.Pose` name  (Default value = True)
        :param method: pass the name of a :class:`.Pose` method to interactively display. Keyword arguments to interactive() will be passed through (Default value = None)
        :param function: pass a callable which will be called as `function(pose)`

        """

        from ipywidgets import (
            interactive,
            BoundedIntText,
            Checkbox,
            interactive_output,
            HBox,
            GridBox,
            Layout,
            VBox,
        )
        from IPython.display import display
        from pprint import pprint

        if method:

            def widget(i):
                pose = self[i]
                if print_name:
                    print(repr(pose))
                value = getattr(pose, method)(**kwargs)
                if value:
                    display(value)

            return interactive(
                widget,
                i=BoundedIntText(
                    value=0,
                    min=0,
                    max=len(self) - 1,
                    step=1,
                    description="Pose:",
                    disabled=False,
                ),
            )

        elif function:

            def widget(i):
                pose = self[i]
                if print_name:
                    display(pose)
                function(pose)

            return interactive(
                widget,
                i=BoundedIntText(
                    value=0,
                    min=0,
                    max=len(self) - 1,
                    step=1,
                    description="Pose:",
                    disabled=False,
                ),
            )

        else:

            a = BoundedIntText(
                value=0,
                min=0,
                max=len(self) - 1,
                step=1,
                description=f"Pose (/{len(self)}):",
                disabled=False,
            )

            b = Checkbox(description="Name", value=True)
            c = Checkbox(description="Summary", value=False)
            h = Checkbox(description="Tags", value=False)
            i = Checkbox(description="Subsites", value=False)
            d = Checkbox(description="2D (Comp.)", value=False)
            e = Checkbox(description="2D (Pose)", value=False)
            f = Checkbox(description="3D", value=True)
            g = Checkbox(description="Metadata", value=False)

            ui1 = GridBox(
                [b, c, d, h],
                layout=Layout(grid_template_columns="repeat(4, 100px)"),
            )
            ui2 = GridBox(
                [e, f, g, i],
                layout=Layout(grid_template_columns="repeat(4, 100px)"),
            )
            ui = VBox([a, ui1, ui2])

            def widget(
                i,
                name=True,
                summary=True,
                grid=True,
                draw2d=True,
                draw=True,
                tags=True,
                subsites=True,
                metadata=True,
            ):
                pose = self[i]
                if name:
                    print(repr(pose))

                if summary:
                    pose.summary(metadata=False, tags=False, subsites=False)
                if tags:
                    print(pose.tags)
                if subsites:
                    print(pose.subsites)
                if grid:
                    pose.grid()
                if draw2d:
                    pose.draw2d()
                if draw:
                    pose.draw()
                if metadata:
                    mrich.title("Metadata:")
                    pprint(pose.metadata)

            out = interactive_output(
                widget,
                {
                    "i": a,
                    "name": b,
                    "summary": c,
                    "grid": d,
                    "draw2d": e,
                    "draw": f,
                    "metadata": g,
                    "tags": h,
                    "subsites": i,
                },
            )

            display(ui, out)

    def summary(self) -> None:
        """Print a summary of this pose set"""
        mrich.header("PoseSet()")
        mrich.var("#poses", len(self))
        mrich.var("#compounds", self.num_compounds)
        mrich.var("tags", self.tags)

    def draw(self) -> None:
        """Render this pose set with Py3Dmol"""

        from molparse.rdkit import draw_mols

        mols = [p.mol for p in self]

        drawing = draw_mols(mols)
        # display(drawing)

    def grid(self) -> None:
        """Draw a grid of all contained molecules"""
        from molparse.rdkit import draw_grid

        data = [(p.name, p.compound.mol) for p in self]

        mols = [d[1] for d in data]
        labels = [d[0] for d in data]

        drawing = draw_grid(mols, labels=labels)
        display(drawing)

    def subsite_summary(self) -> "pd.DataFrame":
        """Print a table counting poses by subsite"""

        from pandas import DataFrame

        sql = f"""
        SELECT subsite_name, COUNT(DISTINCT subsite_tag_pose) FROM subsite
        INNER JOIN subsite_tag
        ON subsite_id = subsite_tag_ref
        WHERE subsite_tag_pose IN {self.str_ids}
        GROUP BY subsite_name
        """

        cursor = self.db.execute(sql)

        df = DataFrame([dict(subsite=name, num_poses=count) for name, count in cursor])

        df = df.set_index("subsite")

        df = df.sort_values(by="num_poses", ascending=False)

        mrich.print(df)

        return df

    ### PRIVATE

    def _delete(self, *, force: bool = False) -> None:
        """Delete poses in this set"""

        if not force:
            mrich.warning("Deleting Poses is risky! Set force=True to continue")
            return

        # delete the poses in this set
        self.db.delete_where(table=self.table, key=f"pose_id IN {self.str_ids}")

        # check for other references to this pose
        self.db.delete_where(table="tag", key=f"tag_pose IN {self.str_ids}")
        self.db.delete_where(
            table="inspiration", key=f"inspiration_original IN {self.str_ids}"
        )
        self.db.delete_where(
            table="inspiration", key=f"inspiration_derivative IN {self.str_ids}"
        )
        self.db.delete_where(
            table="subsite_tag", key=f"subsite_tag_pose IN {self.str_ids}"
        )
        self.db.delete_where(
            table="interaction", key=f"interaction_pose IN {self.str_ids}"
        )

        self.db.execute(
            f"""
            UPDATE pose
            SET pose_reference = NULL
            WHERE pose_id IN {self.str_ids}
        """
        )

    ### DUNDERS

    def __str__(self):
        """Unformatted string representation"""
        if self.name:
            s = f"{self.name}: "
        else:
            s = ""

        s += "{" f"P × {len(self)}" "}"

        return s

    def __repr__(self) -> str:
        """ANSI Formatted string representation"""
        return f"{mcol.bold}{mcol.underline}{self}{mcol.unbold}{mcol.ununderline}"

    def __rich__(self) -> str:
        """Rich Formatted string representation"""
        return f"[bold underline]{self}"

    def __len__(self) -> int:
        """The number of poses in this set"""
        return len(self.indices)

    def __iter__(self):
        """Iterate through poses in this set"""
        return iter(self.db.get_pose(id=i) for i in self.indices)

    def __getitem__(
        self,
        key: int | slice,
    ) -> "Pose | PoseSet":
        """Get poses or subsets thereof from this set

        :param key: integer index or slice of indices

        """
        match key:

            case int():
                try:
                    index = self.indices[key]
                except IndexError:
                    mrich.error(f"list index out of range: {key=} for {self}")
                    raise
                return self.db.get_pose(id=index)

            case slice():
                ids = self.indices[key]
                return PoseSet(self.db, ids)

            case _:
                raise NotImplementedError

    def __add__(
        self,
        other: "PoseSet",
    ) -> "PoseSet":
        """Add a :class:`.PoseSet` to this set"""
        if isinstance(other, PoseSet):
            return PoseSet(self.db, self.ids + other.ids, sort=False)
        elif isinstance(other, Pose):
            return PoseSet(self.db, self.ids + [other.id], sort=False)
        else:
            raise NotImplementedError

    def __sub__(
        self,
        other: "PoseSet",
    ) -> "PoseSet":
        """Substract a :class:`.PoseSet` from this set"""
        match other:
            case PoseSet():
                ids = set(self.ids) - set(other.ids)
                return PoseSet(self.db, ids, sort=False)
            case int():
                # assert other in set(self.ids)
                return PoseSet(self.db, [i for i in self.ids if i != other], sort=False)

    def __call__(
        self,
        *,
        tag: str = None,
        target: int = None,
        subsite: int = None,
    ) -> "PoseSet":
        """Filter poses by a given tag, Subsite ID, or target ID. See :meth:`.PoseSet.get_by_tag`, :meth:`.PoseSet.get_by_target`, amd :meth:`.PoseSet.get_by_subsite`"""

        if tag:
            return self.get_by_tag(tag)
        elif target:
            return self.get_by_target(id=target)
        elif subsite:
            return self.get_by_subsite(id=subsite)
        else:
            raise NotImplementedError



import mcol

import mrich


class Interaction:
    """A :class:`.Interaction` represents an interaction between an rdkit Feature on a :class:`.Pose` and a :class:`.Feature` on the protein :class:`.Target`.

    .. attention::

            :class:`.Interaction` objects should not be created directly. Instead use :meth:`.Pose.interactions`, or :meth:`.PoseSet.interactions` methods.

    """

    def __init__(
        self,
        db: "Database",
        id: int,
        feature_id: int,
        pose_id: int,
        type: str,
        family: str,
        atom_ids: str,
        prot_coord: str,
        lig_coord: str,
        distance: float,
        angle: float,
        energy: float | None,
        table: str = "interaction",
    ) -> None:

        import json

        # from interaction table
        self._id = id
        self._feature_id = feature_id
        self._pose_id = pose_id
        self._type = type
        self._family = family
        self._atom_ids = json.loads(atom_ids)
        self._prot_coord = json.loads(prot_coord)
        self._lig_coord = json.loads(lig_coord)
        self._distance = distance
        self._angle = angle
        self._energy = energy

        # placeholders
        self._pose = None
        self._feature = None
        self._table = table

        self._db = db

    ### PROPERTIES

    @property
    def id(self) -> int:
        """Returns the interaction's database ID"""
        return self._id

    @property
    def table(self) -> str:
        """Returns the name of the :class:`.Database` table"""
        return self._table

    @property
    def db(self) -> "Database":
        """Returns a pointer to the parent database"""
        return self._db

    @property
    def family(self) -> str:
        """The Feature family"""
        return self._family

    @property
    def pose_id(self) -> int:
        """Returns the associated :class:`.Pose`'s database ID"""
        return self._pose_id

    @property
    def pose(self) -> "Pose":
        """Returns the associated :class:`.Pose`'s object"""
        if not self._pose:
            self._pose = self.db.get_pose(id=self.pose_id)
        return self._pose

    @property
    def feature_id(self) -> int:
        """Returns the associated :class:`.Feature`'s database ID"""
        return self._feature_id

    @property
    def feature(self) -> "Feature":
        """Returns the associated :class:`.Feature`'s object"""
        if not self._feature:
            self._feature = self.db.get_feature(id=self.feature_id)
        return self._feature

    @property
    def residue_name(self) -> str:
        """Returns the associated :class:`.Feature`'s residue name"""
        return self.feature.residue_name

    @property
    def residue_number(self) -> int:
        """Returns the associated :class:`.Feature`'s residue number"""
        return self.feature.residue_number

    @property
    def atom_ids(self) -> list[int]:
        """Returns the indices of atoms making up the ligand feature"""
        return self._atom_ids

    @property
    def prot_coord(self) -> list[float]:
        """Returns the cartesian position of the protein :class:`.Feature`"""
        return self._prot_coord

    @property
    def lig_coord(self) -> list[float]:
        """Returns the cartesian position of the ligand feature"""
        return self._lig_coord

    @property
    def distance(self) -> float:
        """Returns the euclidian distance of the interaction"""
        return self._distance

    @property
    def angle(self) -> float | None:
        """Returns the interaction angle (only defined for π-stacking and π-cation interactions)"""
        return self._angle

    @property
    def energy(self) -> float | None:
        """Returns the interaction energy, if defined"""
        return self._energy

    @property
    def family_str(self) -> str:
        """String of the two feature families"""
        return f"{repr(self.feature)} ~ {self.family}"

    @property
    def type(self) -> str:
        """Interaction type string"""
        from molparse.rdkit.features import INTERACTION_TYPES

        return INTERACTION_TYPES[(self.feature.family, self.family)]

    @property
    def description(self) -> str:
        """One line description of this interaction"""
        s = f"{self.type} [{self.feature.chain_res_name_number_str}] {self.distance:.1f} Å"
        if self.angle:
            s += f", {self.angle:.1f} degrees"
        return s

    ### METHODS

    def summary(self) -> None:
        """Print a summary of this interaction's properties"""

        mrich.header(f"Interaction {self.id}")

        mrich.var("feature", self.feature)
        mrich.var("pose", self.pose)
        mrich.var("family", self.family)
        mrich.var("atom_ids", self.atom_ids)
        mrich.var("prot_coord", self.prot_coord)
        mrich.var("lig_coord", self.lig_coord)
        mrich.var("distance", self.distance)
        mrich.var("angle", self.angle)
        mrich.var("energy", self.energy)

    ### DUNDERS

    def __str__(self) -> str:
        """Plain string representation"""
        return f"I{self.id}"

    def __repr__(self) -> str:
        """ANSI Formatted string representation"""
        return f"{mcol.bold}{mcol.underline}{self}{mcol.unbold}{mcol.ununderline}"

    def __rich__(self) -> str:
        """Rich formatted string representation"""
        return f"[bold underline]{self}"



import mcol
import molparse as mp

from rdkit import Chem

import numpy as np

from .tags import TagSet

import pickle

from pathlib import Path

import mrich

from molparse.rdkit.features import (
    FEATURE_FAMILIES,
    COMPLEMENTARY_FEATURES,
    INTERACTION_TYPES,
)

INTERACTION_CUTOFF = {
    "Hydrophobic": 4.5,
    "Hydrogen Bond": 3.5,
    "Electrostatic": 4.5,
    "π-stacking": 6.0,
    "π-cation": 4.5,
}

PI_STACK_MIN_CUTOFF = 3.8
PI_STACK_F2F_CUTOFF = 4.5
PI_STACK_E2F_CUTOFF = 6.0
MUTATION_WARNING_DIST = 15


class Pose:
    """A :class:`.Pose` is a particular conformer of a :class:`.Compound` within a protein environment. A pose will have its own (stereochemical) smiles string, and must have a path to a coordinate file. Poses can have *inspirations* that can be used to trace fragment-derived scaffolds in merges and expansions.

    .. attention::

            :class:`.Pose` objects should not be created directly. Instead use :meth:`.HIPPO.register_pose` or :meth:`.HIPPO.poses`

    """

    _table = "pose"

    def __init__(
        self,
        db: "Database",
        id: int,
        inchikey: str | None,
        alias: str | None,
        smiles: str,
        reference: int,  # another pose
        path: str,
        compound: int,
        target: int,
        mol: Chem.Mol | bytes | None,
        fingerprint: int,
        energy_score: float | None = None,
        distance_score: float | None = None,
        inspiration_score: float | None = None,
        metadata: dict | None = None,
    ):

        self._db = db
        self._id = id
        self._inchikey = inchikey
        self._alias = alias
        self._smiles = smiles
        self._compound_id = compound
        self._target = target
        self._path = path
        self._protein_system = None
        self._energy_score = energy_score
        self._distance_score = distance_score
        self._inspiration_score = inspiration_score

        self._base_ids = None
        self._num_heavy_atoms = None

        self._has_fingerprint = False

        if fingerprint is None:
            self._has_fingerprint = False
        elif not isinstance(fingerprint, int):
            mrich.warning("Legacy fingerprint data format")
            self.has_fingerprint = False
        else:
            self.has_fingerprint = bool(fingerprint)

        # print(f'{self}{metadata=}')
        self._metadata = metadata
        self._tags = None
        self._reference = reference
        self._reference_id = reference
        self._interactions = None

        if isinstance(mol, bytes):
            self._mol = Chem.Mol(mol)
        else:
            self._mol = mol

        self._total_changes = db.total_changes
        self._num_atoms_added_wrt_inspirations = None

    ### FACTORIES

    ### PROPERTIES

    @property
    def db(self) -> "Database":
        """Returns a pointer to the parent database"""
        return self._db

    @property
    def id(self) -> int:
        """Returns the pose's database ID"""
        return self._id

    @property
    def inchikey(self) -> str:
        """Returns the pose's inchikey"""
        if not self._inchikey:
            self.smiles
        return self._inchikey

    @property
    def alias(self) -> str:
        """Returns the pose's alias"""
        return self._alias

    @property
    def name(self) -> str:
        """Returns the pose's name"""
        if n := self.alias:
            return n
        else:
            return self.inchikey

    @alias.setter
    def alias(self, n) -> None:
        """Set the pose's alias"""
        assert isinstance(n, str)
        self._alias = n
        self.db.update(table="pose", id=self.id, key="pose_alias", value=n)

    @inchikey.setter
    def inchikey(self, n) -> None:
        """Set the pose's inchikey"""
        assert isinstance(n, str)
        self._inchikey = n
        self.db.update(table="pose", id=self.id, key="pose_inchikey", value=n)

    @property
    def smiles(self) -> str:
        """Returns the pose's smiles"""
        if not self._smiles:
            from molparse.rdkit import mol_to_smiles
            from rdkit.Chem.inchi import MolToInchiKey

            try:
                mol = self.mol
                self._smiles = mol_to_smiles(mol)
                self.inchikey = MolToInchiKey(mol)
                self.db.update(
                    table="pose", id=self.id, key="pose_smiles", value=self._smiles
                )
            except InvalidMolError:
                mrich.warning(f"Taking smiles from {self.compound}")
                self._smiles = self.compound.smiles
        return self._smiles

    @property
    def target(self) -> "Target":
        """Returns the pose's associated target"""
        if isinstance(self._target, int):
            self._target = self.db.get_target(id=self._target)
        return self._target

    @property
    def compound_id(self) -> int:
        """Returns the pose's associated compound ID"""
        return self._compound_id

    @property
    def compound(self) -> "Compound":
        """Returns the pose's associated compound"""
        return self.get_compound()

    @property
    def path(self) -> str:
        """Returns the pose's path"""
        return self._path

    @property
    def reference(self) -> "Pose":
        """Returns the pose's protein reference (another pose)"""
        if isinstance(self._reference, int):
            self._reference = self.db.get_pose(id=self._reference)
        return self._reference

    @property
    def reference_id(self) -> int:
        """Returns the pose's protein reference ID"""
        return self._reference_id

    @reference.setter
    def reference(self, p):
        """Set the pose's reference"""
        if not isinstance(p, int):
            assert p._table == "pose"
            p = p.id
        self._reference = p
        self._reference_id = p
        self.db.update(table="pose", id=self.id, key="pose_reference", value=p)

    @property
    def mol(self) -> "rdkit.Chem.Mol":
        """Returns a pose's rdkit.Chem.Mol"""
        if not self._mol and self.path:

            if self.path.endswith(".pdb"):

                mrich.reading(self.path)

                # mrich.reading(self.path)
                sys = mp.parse(self.path, verbosity=False)

                self.protein_system = sys.protein_system

                sdf_path = list(Path(self.path).parent.glob("*_ligand.sdf"))

                if len(sdf_path) == 1:

                    supplier = Chem.SDMolSupplier(sdf_path[0])
                    mols = [mol for mol in supplier if mol is not None]

                    if len(mols) > 1:
                        mrich.warning(f"Multiple molecules in SDF {self}")

                    self.mol = mols[0]
                    return self._mol

                # look for ligand mol from Fragalysis
                mol_path = list(
                    Path(self.path).parent.glob("*_ligand.mol")
                )  # str(Path(self.path).name).replace('.pdb','_ligand.mol')

                if len(mol_path) == 1:
                    mol_path = mol_path[0]
                    from rdkit.Chem import MolFromMolFile

                    self._mol_path = mol_path.resolve()

                    mol = MolFromMolFile(str(self._mol_path))

                elif len(mol_path) == 0:

                    lig_residues = sys["rLIG"]

                    if not lig_residues:
                        lig_residues = [r for r in sys.residues if r.type == "LIG"]

                    if len(lig_residues) > 1:
                        mrich.warning(f"Multiple ligands in PDB {self}")

                    lig_res = lig_residues[0]

                    if not (mol := lig_res.rdkit_mol):
                        mrich.error(
                            f"[{self}] Error computing RDKit Mol from PDB={self.path}"
                        )

                        print(lig_res.pdb_block)

                        lig_res.plot3d()

                        raise InvalidMolError

                    # clean up bond orders
                    from rdkit.Chem.AllChem import (
                        MolFromSmiles,
                        AssignBondOrdersFromTemplate,
                    )

                    template = MolFromSmiles(self.compound.smiles)
                    try:
                        mol = AssignBondOrdersFromTemplate(template, mol)
                    except Exception as e:
                        mrich.error(
                            f"Exception occured during AssignBondOrdersFromTemplate for {self}.mol"
                        )
                        print(f"template_smiles={self.compound.smiles}")
                        print(f"pdbblock={print(lig_res.pdb_block)}")
                        mrich.error(e)
                        mol = lig_res.rdkit_mol

                else:

                    mrich.warning(
                        f"There are multiple *_ligand.mol files in {Path(self.path).parent}"
                    )

                self.mol = mol

            elif self.path.endswith(".mol"):

                mrich.reading(self.path)

                # mrich.reading(self.path)
                mol = mp.parse(self.path, verbosity=False)

                if not mol:
                    mrich.error(
                        f"[{self}] Error computing RDKit Mol from .mol={self.path}"
                    )

                    raise InvalidMolError

                self.mol = mol

            else:

                raise NotImplementedError

            if not mol:
                mrich.error(f"Could not parse {self}.path={self.path}")

        return self._mol

    @mol.setter
    def mol(self, m):
        """Set the pose's rdkit.Chem.Mol"""
        assert m
        from .tools import sanitise_mol

        self._mol = sanitise_mol(m)
        self.db.update(table="pose", id=self.id, key="pose_mol", value=m.ToBinary())

    @property
    def protonated_mol(self) -> "rdkit.Chem.Mol":
        """Guess hydrogen positions"""
        from rdkit.Chem import AllChem

        mol = self.mol
        protonated_mol = Chem.AddHs(mol)
        try:
            protonated_mol = AllChem.ConstrainedEmbed(protonated_mol, mol)
        except Exception as e:
            mrich.error("Error while embedding protonated molecule")
            mrich.error(e)
            return mol
        return protonated_mol

    @property
    def protein_system(self) -> "molparse.System":
        """Returns the pose's protein molparse.System"""
        if self._protein_system is None and self.path.endswith(".pdb"):
            # mrich.debug(f'getting pose protein system {self}')
            self.protein_system = mp.parse(self.path, verbosity=False).protein_system
        return self._protein_system

    @protein_system.setter
    def protein_system(self, a):
        """Sets the pose's protein molparse.System"""
        self._protein_system = a

    @property
    def complex_system(self) -> "molparse.System":

        if self.has_complex_pdb_path:
            return mp.parse(self.path, verbosity=False)

        elif self.reference:

            # construct from .mol and reference

            system = self.reference.protein_system.copy()

            system.name = (
                f"{self.target.name}_{self.reference.name}_{self.compound.name}"
            )

            from molparse.rdkit import mol_to_AtomGroup

            ligand = mol_to_AtomGroup(self.mol)

            for atom in ligand.atoms:
                system.add_atom(atom)

            return system

        else:

            raise NotImplementedError

    @property
    def has_complex_pdb_path(self) -> bool:
        """Does this pose have a PDB file?"""
        return self.path.endswith(".pdb")

    @property
    def metadata(self) -> "MetaData":
        """Returns the pose's metadata"""
        if self._metadata is None:
            self._metadata = self.db.get_metadata(table="pose", id=self.id)
        return self._metadata

    @property
    def has_fingerprint(self) -> bool:
        """Does the pose have a fingerprint?"""
        return self._has_fingerprint

    @has_fingerprint.setter
    def has_fingerprint(self, fp):
        self.set_has_fingerprint(fp)

    @property
    def tags(self) -> "TagSet":
        """Returns the pose's tags"""
        if not self._tags:
            self._tags = self.get_tags()
        return self._tags

    @property
    def inspirations(self) -> "PoseSet":
        """Returns the pose's inspirations"""
        return self.get_inspirations()

    @property
    def derivatives(self) -> "PoseSet":
        """Returns the pose's derivatives"""
        return self.get_derivatives()

    @property
    def features(self) -> "list[molparse.rdkit.Feature]":
        """Returns the pose's features"""
        return mp.rdkit.features_from_mol(self.mol)

    @property
    def dict(self) -> dict:
        """Serialised dictionary representing the pose"""
        return self.get_dict()

    @property
    def table(self) -> str:
        """Get the name of the database table"""
        return self._table

    @property
    def num_heavy_atoms(self) -> int:
        """Number of heavy atoms"""
        if not self._num_heavy_atoms:
            self._num_heavy_atoms = self.db.get_compound_computed_property(
                "num_heavy_atoms", self.compound_id
            )
        return self._num_heavy_atoms

    @property
    def num_atoms_added(self) -> int:
        """Calculate the number of atoms added relative to the base or inspirations"""
        if self.num_bases == 1:
            return self.num_atoms_added_wrt_bases
        else:
            return self.num_atoms_added_wrt_inspirations

    @property
    def num_atoms_added_wrt_bases(self) -> int | list[int] | None:
        """Calculate the number of atoms added relative to the base"""
        return self.compound.num_atoms_added

    @property
    def num_atoms_added_wrt_inspirations(self) -> int | None:
        """Calculate the number of atoms added relative to its inspirations"""

        if self._num_atoms_added_wrt_inspirations is None or self._db_changed:

            sql = f"""
            WITH inspirations AS (
                SELECT SUM(mol_num_hvyatms(compound_mol)) AS sum, inspiration_derivative FROM inspiration
                INNER JOIN pose ON inspiration_original = pose_id
                INNER JOIN compound ON pose_compound = compound_id
                WHERE inspiration_derivative = {self.id}
            )
            SELECT mol_num_hvyatms(compound_mol) - sum FROM inspirations
            INNER JOIN pose ON inspiration_derivative = pose_id
            INNER JOIN compound ON compound_id = pose_compound
            """

            (diff,) = self.db.execute(sql).fetchone()

            self._num_atoms_added_wrt_inspirations = diff

        return self._num_atoms_added_wrt_inspirations

    @property
    def num_bases(self) -> int:
        """Get the number of base scaffolds"""
        return len(self.base_ids)

    @property
    def base_ids(self) -> list[int] | None:
        """Get the base :class:`.Compound` IDs"""
        if self._base_ids is None:
            records = self.db.select_where(
                table="scaffold",
                query="scaffold_base",
                key="superstructure",
                value=self.compound_id,
                multiple=True,
                none="quiet",
            )
            records = [i for i, in records]
            self._base_ids = records
        return self._base_ids

    @property
    def energy_score(self) -> float | None:
        """Energy score of the Pose (kcal/mol)"""
        return self._energy_score

    @property
    def distance_score(self) -> float | None:
        """Distance score of the Pose (w.r.t. its inspirations), in Angstroms"""
        return self._distance_score

    @property
    def inspiration_score(self) -> float | None:
        """inspiration score of the Pose in range 0.00-1.00"""
        return self._inspiration_score

    @property
    def interactions(self) -> "InteractionSet":
        """Get a :class:`.InteractionSet` for this :class:`.Pose`"""
        if not self._interactions:
            from .iset import InteractionSet

            self._interactions = InteractionSet.from_pose(self)
        return self._interactions

    @property
    def classic_fingerprint(self) -> dict:
        """Classic HIPPO fingerprint dictionary, mapping protein :class:`.Feature` ID's to the number of corresponding ligand features (from any :class:`.Pose`)"""
        return self.interactions.classic_fingerprint

    @property
    def subsites(self):

        from .subsite import SubsiteTag

        records = self.db.select_where(
            table="subsite_tag",
            key="pose",
            value=self.id,
            multiple=True,
            query="subsite_tag_id, subsite_tag_ref",
            none="quiet",
        )

        if not records:
            return None

        subsite_tags = []
        for record in records:
            id, ref = record
            subsite_tag = SubsiteTag(db=self.db, id=id, subsite_id=ref, pose_id=self.id)
            subsite_tags.append(subsite_tag)

        return subsite_tags

    @property
    def _db_changed(self) -> bool:
        """Has the database changed?"""
        if self._total_changes != self.db.total_changes:
            self._total_changes = self.db.total_changes
            return True
        return False

    @property
    def mol_path(self):
        path = Path(self.path)
        if path.name.endswith(".pdb"):
            mol_path = path.parent / path.name.replace("_hippo.pdb", ".pdb").replace(
                ".pdb", "_ligand.mol"
            )
            if not mol_path.exists():
                mol_path = path.parent / path.name.replace(
                    "_hippo.pdb", ".pdb"
                ).replace(".pdb", "_ligand.sdf")
                if not mol_path.exists():
                    mrich.error("Could not find ligand mol/sdf:", mol_path)
                    return None
            return mol_path
        elif path.name.endswith(".mol"):
            return path
        else:
            raise NotImplementedError

    @property
    def apo_path(self):
        path = Path(self.path)
        if path.name.endswith(".pdb"):
            apo_path = path.parent / path.name.replace("_hippo.pdb", ".pdb").replace(
                ".pdb", "_apo-desolv.pdb"
            )
            if not apo_path.exists():
                return None
            return apo_path
        else:
            raise NotImplementedError

    ### METHODS

    def score_inspiration(
        self,
        debug: bool = False,
        draw: bool = False,
        return_all: bool = False,
    ) -> float:
        """Score how well this Pose recapitulates the pharmacophoric features of its inspirations.

        :param debug: Increased verbosity for debugging (Default value = False)
        :param draw: Render each inspiration pose with it's features, the derivative with the combined features of the inspirations, and the derivative with it's features. (Default value = False)

        """

        # from molparse.rdkit import SuCOS_score
        from mucos import MuCOS_score

        multi_sucos = MuCOS_score(
            self.inspirations.mols,
            self.mol,
            print_scores=debug,
            draw=draw,
            return_all=return_all,
        )

        if debug:
            mrich.var("energy_score", self.energy_score)
            mrich.var("distance_score", self.distance_score)

            for inspiration in self.inspirations:
                mrich.var(
                    f"{inspiration} SuCOS",
                    MuCOS_score(inspiration.mol, self.mol, print_scores=debug),
                )

            mrich.var(f"multi SuCOS", multi_sucos)

        return multi_sucos

    def get_compound(self) -> "Compound":
        """Get the :class:`.Compound` that this pose is a conformer of"""
        return self.db.get_compound(id=self._compound_id)

    def get_tags(self) -> "TagSet":
        """Get this Pose's tags"""
        tags = self.db.select_where(
            query="tag_name",
            table="tag",
            key="pose",
            value=self.id,
            multiple=True,
            none="quiet",
        )
        return TagSet(self, {t[0] for t in tags})

    def get_inspiration_ids(self) -> list[int]:
        """Get the :class:`.Pose` IDs of this pose's inspirations"""
        inspirations = self.db.select_where(
            query="inspiration_original",
            table="inspiration",
            key="derivative",
            value=self.id,
            multiple=True,
            none="quiet",
        )
        if not inspirations:
            return None
        return set([v for v, in inspirations])

    def get_derivative_ids(self) -> list[int]:
        """Get the :class:`.Pose` IDs of this pose's derivatives"""
        derivatives = self.db.select_where(
            query="inspiration_derivative",
            table="inspiration",
            key="original",
            value=self.id,
            multiple=True,
            none="quiet",
        )
        if not derivatives:
            return None
        return set([v for v, in derivatives])

    def get_inspirations(self) -> "PoseSet":
        """Get a :class:`.PoseSet` of this pose's inspirations"""
        if not (inspirations := self.get_inspiration_ids()):
            return None

        from .pset import PoseSet

        return PoseSet(self.db, indices=inspirations)

    def get_derivatives(self) -> "PoseSet":
        """Get a :class:`.PoseSet` of this pose's derivatives"""
        if not (derivatives := self.get_derivative_ids()):
            return None

        from .pset import PoseSet

        return PoseSet(self.db, indices=derivatives)

    def get_dict(
        self,
        mol: bool = False,
        inspirations: bool | str = True,
        subsites: bool | str = True,
        reference: bool | str = True,
        metadata: bool = True,
        duplicate_name: str | bool = False,
        sanitise_null_metadata_values: bool = False,
        skip_metadata: list[str] | None = None,
        sanitise_tag_list_separator: str | None = None,
        sanitise_metadata_list_separator: str | None = ";",
        tags: bool = True,
    ) -> dict:
        """Returns a dictionary representing this Pose. Arguments:

        :param mol: Include a ``rdkit.Chem.Mol`` in the output?  (Default value = False)
        :param inspirations: Include inspirations? ``[True, False, 'names']`` Specify ``names`` to format as a comma-separated string (Default value = True)
        :param subsites: Include subsites? ``[True, False, 'names']`` Specify ``names`` to format as a comma-separated string (Default value = True)
        :param reference: Include reference? ``[True, False, 'name']`` Specify ``name`` to include the :class:`.Pose` name rather than it's ID (Default value = True)
        :param metadata: Include metadata? (Default value = True)
        :param duplicate_name: Specify the name of a new column duplicating the pose name column  (Default value = False)
        :param tags: bool: Include tags? (Default value = True)

        """

        skip_metadata = skip_metadata or []

        serialisable_fields = [
            "id",
            "inchikey",
            "alias",
            "name",
            "smiles",
            "path",
            "distance_score",
            "energy_score",
            "inspiration_score",
        ]

        data = {}
        for key in serialisable_fields:
            data[key] = getattr(self, key)

        if duplicate_name:
            assert isinstance(duplicate_name, str)
            data[duplicate_name] = data["name"]

        if mol:
            try:
                data["mol"] = self.mol
            except InvalidMolError:
                data["mol"] = None

        data["compound"] = self.compound.name
        data["compound_id"] = self.compound.id
        data["target"] = self.target.name

        if tags:
            data["tags"] = self.tags
            if sanitise_tag_list_separator:
                data["tags"] = sanitise_tag_list_separator.join(data["tags"])

        if inspirations == "names":
            if not self.inspirations:
                data["inspirations"] = None
            else:
                data["inspirations"] = ",".join([p.name for p in self.inspirations])
        elif inspirations:
            data["inspirations"] = self.inspirations

        if subsites == "names":
            if not (sites := self.subsites):
                data["subsites"] = None
            else:
                data["subsites"] = ",".join([p.name for p in sites])
        elif subsites:
            data["subsites"] = self.subsites

        if reference == "name":
            if not self.reference:
                data["reference"] = ""
            else:
                data["reference"] = self.reference.name
        elif reference:
            data["reference"] = self.reference

        if metadata and (metadict := self.metadata):
            for key in metadict:

                value = metadict[key]

                if key in skip_metadata:
                    continue

                if (
                    sanitise_null_metadata_values
                    and isinstance(value, str)
                    and not value
                ):
                    value = None

                elif sanitise_metadata_list_separator and isinstance(value, list):

                    new_values = []

                    for v in value:
                        if (
                            sanitise_null_metadata_values
                            and isinstance(v, str)
                            and not v
                        ):
                            v = None

                        else:
                            v = str(v)

                        new_values.append(v)

                    value = sanitise_metadata_list_separator.join(new_values)

                data[key] = value

        return data

    def add_subsite(self, name: str, commit: bool = True) -> "SubsiteTag":
        """Tag this pose with a protein subsite

        :param name: the name of the subsite
        :param commit: commit the insertion to the database
        :returns: :class:`.SubsiteTag`

        """

        id = self.db.insert_subsite_tag(pose_id=self.id, name=name, commit=commit)

        if not id:
            return None

        return self.db.get_subsite_tag(id=id)

    def calculate_interactions(
        self,
        resolve: bool = True,
        distance_padding: float = 0.0,
        angle_padding: float = 0.0,
        force: bool = False,
        debug: bool = False,
        commit: bool = True,
        mutation_warnings: bool = True,
    ) -> None:
        """Enumerate all valid interactions between this ligand and the protein

        :param resolve: Cull duplicate / less-significant interactions
        :param distance_padding: Apply a padding in Angstrom to all distance cutoffs
        :param angle_padding: Apply a padding in degrees to all angle cutoffs
        :param force: Force a recalculation even if the pose has already been fingerprinted
        :param debug: Increase verbosity for debugging
        :param commit: commit the changes to the database (Default value = True)
        :param mutation_warnings: warn when there has been a mutation in the protein (Default value = True)

        """

        if not self.has_fingerprint or force:

            def norm(coords):
                import numpy as np

                coords = np.array(coords).T
                cov = np.cov(coords)
                eig = np.linalg.eig(cov)
                vec = eig[1][:, 0]
                return vec

            def unit_vector(vector):
                """Returns the unit vector of the vector."""
                return vector / np.linalg.norm(vector)

            def angle_between(v1, v2):
                """Returns the angle in radians between vectors 'v1' and 'v2'::

                >>> angle_between((1, 0, 0), (0, 1, 0))
                1.5707963267948966
                >>> angle_between((1, 0, 0), (1, 0, 0))
                0.0
                >>> angle_between((1, 0, 0), (-1, 0, 0))
                3.141592653589793
                """
                v1_u = unit_vector(v1)
                v2_u = unit_vector(v2)
                a = 180 * np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0)) / np.pi

                if a > 90:
                    a = 180 - a
                return a

            # calculate interactions...

            ### clear old interactions

            self.db.delete_where(
                table="interaction", key="pose", value=self.id, commit=commit
            )
            self.set_has_fingerprint(False, commit=commit)
            self._interactions = None

            ### create temporary table

            if "temp_interaction" in self.db.table_names:
                mrich.warning("Deleting existing temp_interaction table")
                self.db.execute("DROP TABLE temp_interaction")

            self.db.create_table_interaction(table="temp_interaction", debug=False)

            ### load the ligand structure

            if self.path.endswith(".pdb"):
                from molparse import parse

                protein_system = self.protein_system
                if not self.protein_system:
                    protein_system = parse(self.path, verbosity=False).protein_system

            elif self.path.endswith(".mol") and self.reference:
                protein_system = self.reference.protein_system

            else:
                mrich.debug("Unsupported: Pose.calculate_interactions()")
                raise NotImplementedError(f"{self}, {self.reference=}, {self.path=}")

            assert protein_system

            if not self.mol:
                mrich.error(f"Could not read molecule for {self}")
                return

            ### get features

            comp_features = self.features
            protein_features = self.target.calculate_features(
                protein_system, reference_id=self.reference_id
            )

            ### organise ligand features by family
            comp_features_by_family = {}
            for family in FEATURE_FAMILIES:
                comp_features_by_family[family] = [
                    f for f in comp_features if f.family == family
                ]

            ### protein chain names
            chains = protein_system.chain_names

            mutation_warnings = set()
            mutation_count = 0

            # loop over protein features
            for prot_feature in protein_features:

                # skip chains that aren't present
                if prot_feature.chain_name not in chains:
                    continue

                prot_family = prot_feature.family

                prot_residue = protein_system.get_chain(
                    prot_feature.chain_name
                ).residues[f"n{prot_feature.residue_number}"]

                if not prot_residue:
                    continue

                if prot_residue.name != prot_feature.residue_name:
                    com = prot_residue.centre_of_mass()
                    if any(
                        np.linalg.norm(com - cf.position) < MUTATION_WARNING_DIST
                        for cf in comp_features
                    ):
                        mutation_warnings.add(
                            f"{prot_residue.name} {prot_residue.number} -> {prot_feature.residue_name} {prot_feature.residue_number}"
                        )
                        mutation_count += 1
                    continue

                ### calculate protein coordinate
                prot_atoms = []
                for atom_name in prot_feature.atom_names.split(" "):
                    atom = prot_residue.get_atom(atom_name, verbosity=0)
                    if atom:
                        prot_atoms.append(atom)

                prot_coords = [a.np_pos for a in prot_atoms if a is not None]

                if not prot_coords:
                    # mrich.warning("Skipping feature with no atoms")
                    continue

                prot_coord = np.array(np.sum(prot_coords, axis=0) / len(prot_atoms))

                if prot_family not in COMPLEMENTARY_FEATURES:
                    continue

                complementary_families = COMPLEMENTARY_FEATURES[prot_family]

                # print(prot_family, complementary_families)

                for complementary_family in complementary_families:

                    interaction_type = INTERACTION_TYPES[
                        (prot_family, complementary_family)
                    ]

                    complementary_comp_features = comp_features_by_family[
                        complementary_family
                    ]

                    for lig_feature in complementary_comp_features:

                        distance = np.linalg.norm(lig_feature - prot_coord)
                        angle = None

                        # check distance cutoff
                        if (
                            distance
                            > INTERACTION_CUTOFF[interaction_type] + distance_padding
                        ):
                            continue

                        # special rules for aromatics
                        if interaction_type.startswith("π"):
                            lig_coords = [
                                self.mol.GetConformer().GetAtomPosition(i - 1)
                                for i in lig_feature.atom_numbers
                            ]

                        # special rules for pi-stacking
                        if interaction_type == "π-stacking":

                            # calculate minimum distance
                            min_distance = None
                            for lig_coord in lig_coords:
                                for p_coord in prot_coords:
                                    d = np.linalg.norm(lig_coord - p_coord)
                                    if not min_distance or d < min_distance:
                                        min_distance = d

                            # skip interaction if no atom is within PI_STACK_MIN_CUTOFF
                            if min_distance > PI_STACK_MIN_CUTOFF + distance_padding:
                                if debug:
                                    print(
                                        f"skipping {prot_feature} due to pi-stack min_distance"
                                    )
                                    print(
                                        prot_feature.residue_name,
                                        prot_feature.residue_number,
                                        prot_feature.chain_name,
                                        prot_feature.family,
                                        lig_feature,
                                        distance,
                                    )
                                continue

                            ### angles
                            lig_norm = norm([list(p) for p in lig_coords])
                            prot_norm = norm(prot_coords)
                            angle = angle_between(lig_norm, prot_norm)

                            # Face to Face has more stringent restraints
                            if (
                                angle < 40 - angle_padding
                                and distance > PI_STACK_F2F_CUTOFF + distance_padding
                            ):
                                if debug:
                                    print(prot_feature.res_name_number_family_str)
                                    print(angle, distance)
                                continue

                        # special rules for pi-cation
                        elif interaction_type == "π-cation":

                            # construct vectors
                            if prot_family == "Aromatic":
                                aromatic_norm = norm(prot_coords)
                                cation_vec = lig_feature.position - prot_coord
                            else:
                                aromatic_norm = norm([list(p) for p in lig_coords])
                                cation_vec = prot_coord - lig_feature.position

                            # calculate angle
                            angle = angle_between(aromatic_norm, cation_vec)

                            # skip if angle too large
                            if angle > 30 + angle_padding:
                                if debug:
                                    print(prot_feature.res_name_number_family_str)
                                    print(angle, distance)
                                continue

                        if debug:
                            print(prot_feature, lig_feature)

                        # insert into the Database
                        self.db.insert_interaction(
                            feature=prot_feature.id,
                            pose=self.id,
                            type=interaction_type,
                            family=lig_feature.family,
                            atom_ids=lig_feature.atom_numbers,
                            prot_coord=prot_coord,
                            lig_coord=lig_feature.position,
                            distance=distance,
                            angle=angle,
                            energy=None,
                            commit=commit,
                            table="temp_interaction",
                        )

            if mutation_warnings:
                mrich.warning(
                    f"Skipped {mutation_count} protein features because the residue was mutated:"
                )
                for mutation in mutation_warnings:
                    mrich.warning(mutation)

            if resolve:
                from .iset import InteractionSet

                interactions = InteractionSet.from_pose(self, table="temp_interaction")
                interactions.resolve(debug=debug)
                # self.interactions.resolve(debug=debug, table='temp_interaction')

            ### transfer interactions from temporary table
            self.db.copy_temp_interactions()
            self.set_has_fingerprint(True, commit=commit)

            ### delete temporary table

            self.db.execute("DROP TABLE temp_interaction")

        elif debug:
            mrich.warning(f"{self} is already fingerprinted, no new calculation")

    def calculate_prolif_interactions(
        self,
        return_all: bool = False,
        max_retry: int = 5,
        use_mda: bool = False,
        force: bool = False,
        clear_existing: bool = True,
        debug: bool = False,
        resolve: bool = True,
    ) -> "prolif.Fingerprint":
        """Use ProLIF to populate the interactions table"""

        if not self.has_fingerprint or force:

            ### clear old interactions

            if clear_existing:
                self.db.delete_where(
                    table="interaction", key="pose", value=self.id, commit=False
                )
                self.set_has_fingerprint(False, commit=False)

            ### create temporary table

            table = "temp_interaction"

            if "temp_interaction" in self.db.table_names:
                mrich.warning("Deleting existing temp_interaction table")
                self.db.execute("DROP TABLE temp_interaction")

            self.db.create_table_interaction(table="temp_interaction", debug=False)

            if not clear_existing:
                self.db.copy_interactions_to_temp(pose_id=self.id)

            # clear cached InteractionSet
            self._interactions = None

            import prolif as plf
            from tempfile import NamedTemporaryFile
            from .prolif import parse_prolif_interactions
            from MDAnalysis import Universe
            import logging

            mdanalysis_logger = logging.getLogger("MDAnalysis")
            mdanalysis_logger.setLevel(logging.WARNING)

            ## prepare inputs

            # decide if MDA is needed
            unprotonated_sys = self.protein_system
            residue_names = set(r.name for r in unprotonated_sys.residues)
            nonstandard = ["HID", "HIE", "HSE", "HSD", "HSP"]
            if any(r in residue_names for r in nonstandard):
                mrich.debug("Using MDA")
                use_mda = True

            # protonated protein
            for i in range(max_retry):
                try:
                    protonated_sys, protein_file = unprotonated_sys.add_hydrogens(
                        return_file=True
                    )

                    if use_mda:
                        with mrich.loading("Creating MDAnalysis.Universe"):
                            universe = Universe(protein_file.name)
                            protein_mol = plf.Molecule.from_mda(universe)
                    else:
                        with mrich.loading("Creating protein rdkit.Chem.Mol"):
                            rdkit_prot = Chem.MolFromPDBFile(
                                protein_file.name, removeHs=False
                            )
                            protein_mol = plf.Molecule(rdkit_prot)

                    break

                except Exception as e:
                    mrich.warning(
                        f"Could not create satisfactory protein molecule, attempts = {i+1}/{max_retry}"
                    )
                    mrich.warning(e)
                    use_mda = True
                    continue
            else:
                mrich.error(
                    f"Tried {max_retry} times to create protein molecule and failed"
                )
                return None

            # ligand
            ligand_file = NamedTemporaryFile(mode="w+t", suffix=".sdf")
            writer = Chem.SDWriter(ligand_file.name)
            writer.write(self.protonated_mol)
            writer.close()
            ligand_iterable = plf.sdf_supplier(ligand_file.name)

            ## run prolif

            fp = plf.Fingerprint(count=True)
            fp.run_from_iterable(ligand_iterable, protein_mol, progress=False, n_jobs=1)

            ## parse outputs and insert Feature and Interaction records
            parse_prolif_interactions(
                self, fp, protonated_sys, debug=debug, table=table
            )

            if resolve:
                from .iset import InteractionSet

                interactions = InteractionSet.from_pose(self, table="temp_interaction")
                interactions.resolve(debug=debug)

            self.db.copy_temp_interactions()
            self.set_has_fingerprint(True, commit=True)

            ### delete temporary table

            self.db.execute("DROP TABLE temp_interaction")

            ## close files
            protein_file.close()
            ligand_file.close()

            if return_all:
                return fp, ligand_iterable, protein_mol

    def calculate_classic_fingerprint(
        self,
        debug: bool = False,
    ) -> dict:
        """Calculate the pose's interaction fingerprint"""

        if self.path.endswith(".pdb"):

            import molparse as mp

            protein_system = self.protein_system
            if not self.protein_system:
                # mrich.reading(self.path)
                protein_system = mp.parse(self.path, verbosity=False).protein_system

        elif self.path.endswith(".mol") and self.reference:

            # mrich.debug('fingerprint from .mol and reference pose')
            protein_system = self.reference.protein_system

        else:

            mrich.debug("Unsupported: Pose.calculate_fingerprint()")
            raise NotImplementedError(f"{self.reference=}, {self.path=}")

        assert protein_system

        if not self.mol:
            return

        comp_features = self.features

        comp_features_by_family = {}
        for family in FEATURE_FAMILIES:
            comp_features_by_family[family] = [
                f for f in comp_features if f.family == family
            ]

        # protein_features = self.target.features
        # if not protein_features:
        protein_features = self.target.calculate_features(protein_system)

        fingerprint = {}

        chains = protein_system.chain_names

        for prot_feature in protein_features:

            if prot_feature.chain_name not in chains:
                continue

            prot_family = prot_feature.family

            prot_residue = protein_system.get_chain(prot_feature.chain_name).residues[
                f"n{prot_feature.residue_number}"
            ]

            if not prot_residue:
                continue

            # if prot_feature.residue_number == 77:
            #   mrich.debug(repr(prot_feature))

            if prot_residue.name != prot_feature.residue_name:
                mrich.warning(f"Feature {repr(prot_feature)}")
                continue

            prot_atoms = [
                prot_residue.get_atom(a) for a in prot_feature.atom_names.split(" ")
            ]

            prot_coords = [a.np_pos for a in prot_atoms if a is not None]

            prot_coord = np.array(np.sum(prot_coords, axis=0) / len(prot_atoms))

            complementary_family = COMPLEMENTARY_FEATURES[prot_family]

            complementary_comp_features = comp_features_by_family[complementary_family]

            cutoff = FEATURE_PAIR_CUTOFFS[f"{prot_family} {complementary_family}"]

            valid_features = [
                f
                for f in complementary_comp_features
                if np.linalg.norm(f - prot_coord) <= cutoff
            ]

            if valid_features:
                if debug:
                    mrich.debug(
                        f"PROT: {prot_feature.residue_name} {prot_feature.residue_number} {prot_feature.atom_names}, LIG: #{len(valid_features)} {[f for f in valid_features]}"
                    )
                fingerprint[prot_feature.id] = len(valid_features)

        return fingerprint

        # self.fingerprint = fingerprint

    def draw(
        self,
        inspirations: bool = True,
        protein: bool = False,
        **kwargs,
    ) -> None:
        """Render this pose (and its inspirations)

        :param inspirations: Render the inspirations? (Default value = True)
        :param protein: Render the protein? This wraps :meth:`.Pose.render` (Default value = False)

        """

        if protein:
            self.render(**kwargs)

        from molparse.rdkit import draw_mols

        mols = [self.mol]
        if inspirations and self.inspirations:
            mols += [i.mol for i in self.inspirations]

        draw_mols(mols)

    def draw2d(
        self,
    ) -> None:
        """Draw a 2D drawing of this pose"""
        from rdkit.Chem import MolFromSmiles

        mol = MolFromSmiles(self.smiles)
        display(mol)

    def render(
        self,
        protein="cartoon",
        ligand="stick",
        protein_color="spectrum",
        interactions: bool = True,
        file: str | None = None,
    ) -> None:
        """Render this pose with the protein using py3Dmol

        :param protein: protein representation, default = 'cartoon'
        :param ligand: ligand representation, default = 'stick'
        :param protein_color: color of protein representation, default = 'spectrum'

        """

        from molparse.py3d import render

        sys = self.complex_system

        def make_view(width="640px", height="480px"):

            view = render(
                sys,
                protein=protein,
                ligand=ligand,
                protein_color=protein_color,
                width=width,
                height=height,
            )

            if interactions:
                COLORS = {
                    "Hydrophobic": "green",
                    "Hydrogen Bond": "blue",
                    "π-stacking": "purple",
                    "π-cation": "pink",
                    "Electrostatic": "red",
                }

                iset = self.interactions

                if not iset:
                    return view

                df = iset.df

                residues = set()

                for i, row in df.iterrows():

                    prot_coord = row["prot_coord"]
                    lig_coord = row["lig_coord"]
                    type = row["type"]
                    color = COLORS.get(type, "black")

                    view.addCylinder(
                        {
                            "start": {
                                "x": prot_coord[0],
                                "y": prot_coord[1],
                                "z": prot_coord[2],
                            },
                            "end": {
                                "x": lig_coord[0],
                                "y": lig_coord[1],
                                "z": lig_coord[2],
                            },
                            # 'radius': radius,
                            "color": color,
                        }
                    )

                    residues.add((row["residue_name"], row["residue_number"]))

                for res_name, res_num in residues:
                    res = sys.residues[f"{res_name} n{res_num}"]
                    view.addModel(res.pdb_block, "pdb")
                    view.setStyle({"model": -1}, {ligand: {}})

            return view

        if file:
            view = make_view(width="100%", height="100%")
            html = view._make_html()

            mrich.writing(file)
            with open(file, "w") as f:
                f.write(html)

        view = make_view()
        return view._repr_html_()

    def grid(self) -> None:
        """Draw a grid of this pose with its inspirations"""
        from molparse.rdkit import draw_grid
        from IPython.display import display

        mols = [self.compound.mol]
        labels = [self.plain_repr()]
        if self.inspirations:
            mols += [i.compound.mol for i in self.inspirations]
            labels += [i.plain_repr() for i in self.inspirations]

        display(draw_grid(mols, labels=labels))

    def summary(
        self, metadata: bool = True, tags: bool = True, subsites: bool = True
    ) -> None:
        """Print a summary of this pose

        :param metadata: include metadata (Default value = True)

        """
        if self.alias:
            mrich.header(f"{str(self)}: {self.alias}")
        else:
            mrich.header(f"{str(self)}: {self.inchikey}")
        mrich.var("inchikey", self.inchikey)
        mrich.var("alias", self.alias)
        mrich.var("smiles", self.smiles)
        mrich.var("compound", self.compound)
        mrich.var("path", self.path)
        mrich.var("target", self.target)
        mrich.var("reference", self.reference)
        if tags:
            mrich.var("tags", self.tags)
        if subsites:
            mrich.var("subsites", self.subsites)
        mrich.var("num_heavy_atoms", self.num_heavy_atoms)
        mrich.var("distance_score", self.distance_score)
        mrich.var("energy_score", self.energy_score)
        mrich.var("inspiration_score", self.inspiration_score)
        if inspirations := self.inspirations:
            mrich.var("inspirations", self.inspirations.names)
            mrich.var("num_atoms_added", self.num_atoms_added)
        if metadata:
            mrich.var("metadata", str(self.metadata))

    def showcase(self) -> None:
        """Print and render this pose as if you were using :meth:`.PoseSet.interactive`"""

        self.summary(metadata=False)
        self.grid()
        self.draw()
        from pprint import pprint

        mrich.title("Metadata:")
        pprint(self.metadata)

    def plain_repr(self) -> str:
        """Unformatted detailed string representation"""
        if self.name:
            return f'{self.compound}->{self}: "{self.name}"'
        else:
            return f"{self.compound}->{self}"

    def plot3d(
        self,
        features: bool = False,
        **kwargs,
    ) -> "plotly.graph_objects.Figure":
        """Use Molparse/Plotly to create a 3d figure of this pose

        :param features: include the features in the figure
        :returns: a plotly Figure object

        """

        mol = self.mol

        import molparse as mp

        group = mp.rdkit.mol_to_AtomGroup(mol)

        if features:
            features = self.features

        return mp.go.plot3d(atoms=group.atoms, features=features, **kwargs)

    def set_has_fingerprint(self, fp: bool, commit: bool = True) -> None:
        """Update the database to reflect this pose's has_fingerprint property"""
        assert isinstance(fp, bool)
        self._has_fingerprint = fp
        self.db.update(
            table="pose",
            id=self.id,
            key=f"pose_fingerprint",
            value=int(fp),
            commit=commit,
        )

    def posebusters(self, debug: bool = False) -> bool:
        """Run a posebusters ligand check on this pose's molecule"""

        # use syndirella implementation
        from syndirella.slipper import intra_geometry, flatness

        geometries: Dict = intra_geometry.check_geometry(self.mol, threshold_clash=0.4)
        flat_results: Dict = flatness.check_flatness(self.mol)

        if not geometries["results"]["bond_lengths_within_bounds"]:
            if debug:
                mrich.debug(self, "did not pass bond length checks.")
            return False
        if not geometries["results"]["bond_angles_within_bounds"]:
            if debug:
                mrich.debug(self, "did not pass bond angle checks.")
            return False
        if not geometries["results"]["no_internal_clash"]:
            if debug:
                mrich.debug(self, "did not pass internal clash checks.")
            return False
        if not flat_results["results"]["flatness_passes"]:
            if debug:
                mrich.debug(self, "did not pass flatness checks.")
            return False
        return True

    def to_syndirella(self, out_key: "str | Path") -> "DataFrame":
        """Create syndirella inputs. See :meth:`.PoseSet.to_syndirella`"""
        from .pset import PoseSet

        return PoseSet(self.db, [self.id]).to_syndirella(
            out_key=out_key, separate=False
        )

    ### DUNDERS

    def __str__(self) -> str:
        """Unformatted string representation"""
        return f"P{self.id}"

    def __repr__(self) -> str:
        """ANSI Formatted string representation"""
        return f"{mcol.bold}{mcol.underline}{self.plain_repr()}{mcol.unbold}{mcol.ununderline}"

    def __rich__(self) -> str:
        """Formatted string representation"""
        return f"[bold underline]{self.plain_repr()}"

    def __eq__(self, other: "Pose") -> bool:
        """Compare this pose with another instance"""

        if isinstance(other, int):
            return self.id == other

        return self.id == other.id

    def __add__(
        self,
        other: "Pose | PoseSet",
    ) -> "PoseSet":
        """Add a :class:`.PoseSet` to this pose"""
        from .pset import PoseSet

        if isinstance(other, PoseSet):
            return PoseSet(self.db, [self.id] + other.ids, sort=False)
        elif isinstance(other, Pose):
            return PoseSet(self.db, [self.id, other.id], sort=False)
        else:
            raise NotImplementedError


class InvalidMolError(Exception):
    """Exception to be thrown when the molecule could not be parsed"""

    ...



import mrich


def generate_header(
    pose,
    method,
    ref_url,
    submitter_name,
    submitter_email,
    submitter_institution,
    generation_date: str | None = None,
    extras=None,
    metadata: bool = True,
) -> "Chem.Mol":
    """Generate a header molecule for Fragalysis RHS upload"""

    extras = extras or {}

    from rdkit.Chem.AllChem import EmbedMolecule
    from molparse.rdkit import mol_from_smiles
    from datetime import date

    header = mol_from_smiles(pose.compound.smiles)

    header.SetProp("_Name", "ver_1.2")
    EmbedMolecule(header)

    generation_date = str(generation_date or date.today())

    header.SetProp("ref_url", ref_url)
    header.SetProp("submitter_name", submitter_name)
    header.SetProp("submitter_email", submitter_email)
    header.SetProp("submitter_institution", submitter_institution)
    header.SetProp("generation_date", generation_date)
    header.SetProp("method", method)

    if metadata:
        for k, v in pose.metadata.items():
            header.SetProp(k, str(k))

    for k, v in extras.items():
        header.SetProp(k, str(v))

    return header


def download_target(
    target_name: str,
    target_access_string: str,
    *,
    destination: "str | Path" = ".",
    stack: str = "production",
    unzip: bool = True,
    overwrite: bool = False,
) -> "Path":
    """Download a target from Fragalysis

    :param target_name: Name of the target
    :param destination: where to put the file(s)
    :param stack: Choose the Fragalysis stack from ['production', 'staging']. Defaults to 'production'
    :param unzip: Unzip the download
    :param overwrite: Overwrite existing downloads
    :returns: a pathlib.Path object to the .zip archive or unpacked directory

    """

    import mcol
    import requests
    from pathlib import Path
    import urllib.request

    destination = Path(destination)

    if not destination.exists():
        mrich.writing(destination)
        destination.mkdir(exist_ok=True, parents=True)

    root = STACK_URLS[stack]

    payload = {
        "all_aligned_structures": True,
        "cif_info": False,
        "diff_file": False,
        "event_file": False,
        "file_url": "",
        "map_info": False,
        "metadata_info": True,
        "mtz_info": False,
        "pdb_info": False,
        "proteins": "",
        "sigmaa_file": False,
        "single_sdf_file": True,
        "static_link": False,
        "target_name": target_name,
        "target_access_string": target_access_string,
        "trans_matrix_info": False,
    }

    mrich.print("Requesting download...")

    url = root + "/api/download_structures/"

    mrich.var("url", url)

    response = requests.post(url, json=payload)

    if response.status_code == 200:
        mrich.print("Download is ready.")
    else:
        mrich.error(f"Download request failed: {response.status_code=}")
        mrich.error(response.text)
        return None

    file_url = urllib.request.pathname2url(response.json()["file_url"])

    zip_path = destination / Path(file_url).name

    if zip_path.exists() and not overwrite:
        mrich.warning(f"Using existing {zip_path}")

    else:

        if zip_path.exists():
            mrich.warning(f"Overwriting {zip_path}")

        mrich.writing(zip_path)

        url = f"{root}/api/download_structures/?file_url={file_url}"

        mrich.var("url", url)

        filename, headers = urllib.request.urlretrieve(url, filename=zip_path)

    if not zip_path.exists():
        mrich.error("Download failed")
        return None

    if unzip:
        try:
            import zipfile

            mrich.print(f"Unzipping {zip_path}")

            target_dir = destination / Path(file_url).name.removesuffix(".zip")

            target_dir.mkdir(exist_ok=overwrite)

            mrich.writing(target_dir)
            with zipfile.ZipFile(zip_path, "r") as zip_ref:
                zip_ref.extractall(target_dir)

        except FileExistsError:
            mrich.warning(
                f"Did not unzip as directory {target_dir} exists. Set overwrite=True to override in future"
            )
            unzip = False

    mrich.success(f"Downloaded {mcol.varName}{target_name}{mcol.success} from {stack}")

    if unzip:
        return Path(target_dir)
    else:
        return Path(zip_path)


def parse_observation_longcode(longcode: str) -> dict[str]:
    """Parse a Fragalysis longcode and try to extract the following information:

    - Target name (target)
    - Crystal/dataset code (crystal)
    - Chain letter (chain)
    - Residue number (residue_number)
    - Version number (version)

    :returns: dictionary of the above keys in parentheses
    """

    import re

    match = re.search(
        r"(.*)_([A-z]_[0-9]*_[0-9])_(.*)\+([A-z]\+[0-9]*\+[0-9])_.LIG", longcode
    )

    if not match:
        raise UnsupportedFragalysisLongcodeError(longcode)

    cryst_str, lig_str, _, _ = match.groups()

    chain, residue_number, version = lig_str.split("_")

    residue_number = int(residue_number)
    version = int(version)

    if match := re.search(r"(.*)-(\w[0-9]{4})", cryst_str):

        target_name = match.group(0)
        crystal = match.group(1)

    else:

        target_name = None
        crystal = cryst_str

    return dict(
        target=target_name,
        crystal=crystal,
        chain=chain,
        residue_number=residue_number,
        version=version,
    )


def find_observation_longcode_matches(
    query: str, codes: list[str], debug: bool = False, allow_version_none: bool = False
) -> list[str]:

    dq = parse_observation_longcode(query)

    keys = dq.keys()

    if debug:
        mrich.var("allow_version_none", allow_version_none)
        mrich.var("dq", str(dq))

    matches = []

    for code in codes:

        if code == query:
            if debug:
                mrich.debug("exact match")
            matches.append(code)
            continue

        dc = parse_observation_longcode(code)

        for key in keys:

            if (
                allow_version_none
                and key == "version"
                and (dc[key] is None or dq[key] is None)
            ):
                continue

            if dc[key] != dq[key]:
                break
        else:
            if debug:
                mrich.debug(f"{query} matches {code}")
            matches.append(code)

    if debug:
        mrich.var("#matches", len(matches))

    if len(matches) < 1 and not allow_version_none:
        return find_observation_longcode_matches(query, codes, allow_version_none=True)

    return matches


STACK_URLS = {
    "production": "https://fragalysis.diamond.ac.uk",
    "staging": "https://fragalysis.xchem.diamond.ac.uk",
}


class UnsupportedFragalysisLongcodeError(NotImplementedError): ...



from mout import debug_log

import mrich

"""

Checks
======

- Num heavy atoms difference
- Formula checks
- Num rings difference

"""

SUPPORTED_CHEMISTRY = {
    "Amidation": {
        "heavy_atoms_diff": 1,
        "rings_diff": 0,
        "atomtype": {
            "removed": {"O": 1, "H": 2},
        },
    },
    "Ester_amidation": {
        "heavy_atoms_diff": ">=3",
        "rings_diff": 0,
        "atomtype": {
            "removed": {"O": ">=1", "*": "*"},
        },
    },
    "Williamson_ether_synthesis": {
        "heavy_atoms_diff": 1,
        "rings_diff": 0,
        "atomtype": {
            "removed": {"Ha": 1, "H": 1},  # any halogen
        },
    },
    "N-Boc_deprotection": {
        "heavy_atoms_diff": 7,
        "rings_diff": 0,
        "atomtype": {
            "removed": {"O": 2, "C": 5, "H": 8},
        },
    },
    "TBS_alcohol_deprotection": {
        "heavy_atoms_diff": 7,
        "rings_diff": 0,
        "atomtype": {
            "removed": {"C": 6, "Si": 1, "H": 14},
        },
    },
    "Sp3-sp2_Suzuki_coupling": {
        # "heavy_atoms_diff": 10,
        "heavy_atoms_diff": ">=4",
        "rings_diff": ">=0",
        "atomtype": {
            # "removed": {"C": 6, "O": 2, "B": 1, "Ha": 1, "H": 12},  # any halogen
            "removed": {"C": ">=0", "O": 2, "B": 1, "Ha": 1, "H": ">=2"},  # any halogen
        },
    },
    "Sp2-sp2_Suzuki_coupling": {
        "heavy_atoms_diff": ">=4",
        "rings_diff": ">=0",
        "atomtype": {
            "removed": {"C": ">=0", "O": 2, "B": 1, "Ha": 1, "H": ">=2"},  # any halogen
        },
    },
    "Buchwald-Hartwig_amidation_with_amide-like_nucleophile": {
        "heavy_atoms_diff": 1,
        "rings_diff": 0,
        "atomtype": {
            "removed": {"Ha": 1, "H": 1},
        },
    },
    "Buchwald-Hartwig_amination": {
        "heavy_atoms_diff": 1,
        "rings_diff": 0,
        "atomtype": {
            "removed": {"Ha": 1, "H": 1},
        },
    },
    "Nucleophilic_substitution_with_amine": {
        "heavy_atoms_diff": 1,
        "rings_diff": 0,
        "atomtype": {
            "removed": {"Ha": 1, "H": 1},
        },
    },
    "N-nucleophilic_aromatic_substitution": {
        "heavy_atoms_diff": 1,
        "rings_diff": 0,
        "atomtype": {
            "removed": {"Ha": 1, "H": 1},  # any halogen
        },
    },
    "Reductive_amination": {
        "heavy_atoms_diff": 1,
        "rings_diff": 0,
        "atomtype": {
            "removed": {"O": 1},  # any halogen
        },
    },
    "Mitsunobu_reaction_with_amine_alcohol_and_thioalcohol": {
        "heavy_atoms_diff": 1,
        "rings_diff": 0,
        "atomtype": {
            "removed": {"O": 1, "H": ">=1"},
        },
    },
    "Steglich_esterification": {
        "heavy_atoms_diff": 1,
        "rings_diff": 0,
        "atomtype": {
            "removed": {"O": 1, "H": 2},
        },
    },
    "Benzyl_alcohol_deprotection": {
        "heavy_atoms_diff": 7,
        "rings_diff": 1,
        "atomtype": {
            "removed": {"C": 7, "H": 6},
        },
    },
    "N-Bn_deprotection": {
        "heavy_atoms_diff": 7,
        "rings_diff": 1,
    },
    "Formation_of_urea_from_two_amines": {
        "heavy_atoms_diff": -2,
        "rings_diff": 0,
    },
    "Amide_Schotten-Baumann_with_amine": {
        "heavy_atoms_diff": 1,
        "rings_diff": 0,
    },
    "Nucleophilic_substitution": {
        "heavy_atoms_diff": 1,
        "rings_diff": 0,
    },
}


def check_reaction_types(types: list[str]) -> None:
    """
    Prints a warning if any of the reaction type strings in ``types`` are not in ``SUPPORTED_CHEMISTRY``

    :param types: A list of reaction type strings to check
    """

    for reaction_type in types:
        if reaction_type not in SUPPORTED_CHEMISTRY:
            mrich.error(f"Can't check chemistry of unsupported {reaction_type=}")


def check_chemistry(reaction_type, reactants, product, debug=False):

    if reaction_type not in SUPPORTED_CHEMISTRY:

        mrich.var("reactants", reactants.ids)
        mrich.var("product", product)

        raise UnsupportedChemistryError(f"Unsupported {reaction_type=}")

    assert reactants
    assert product

    CHEMISTRY = SUPPORTED_CHEMISTRY[reaction_type]

    if "heavy_atoms_diff" in CHEMISTRY:
        check = check_count_diff(
            "heavy_atoms", reaction_type, reactants, product, debug=debug
        )
        if not check:
            return False

    if "rings_diff" in CHEMISTRY:
        check = check_count_diff(
            "rings", reaction_type, reactants, product, debug=debug
        )
        if not check:
            return False

    if "atomtype" in CHEMISTRY:
        check = check_atomtype_diff(reaction_type, reactants, product, debug=debug)
        if not check:
            return False

    if debug:
        mrich.success(f"{reaction_type}: All OK")

    return True


def check_count_diff(check_type, reaction_type, reactants, product, debug=False):
    """

    :param check_type:
    :param reaction_type:
    :param reactants:
    :param product:
    :param debug:  (Default value = False)

    """

    # get target value
    diff = SUPPORTED_CHEMISTRY[reaction_type][f"{check_type}_diff"]

    # get attribute name
    attr = f"num_{check_type}"

    # get values
    reac_count = getattr(reactants, attr)
    prod_count = getattr(product, attr)
    if debug:
        mrich.var(f"#{check_type} reactants", reac_count)
    if debug:
        mrich.var(f"#{check_type} product", prod_count)

    # check against target value
    if isinstance(diff, str):

        assert diff.startswith(">="), diff

        diff = int(diff[2:])

        if reac_count - prod_count < diff:
            if debug:
                mrich.error(
                    f"{reaction_type}: #{check_type} {(reac_count - prod_count)=} FAIL"
                )
            return False

        elif debug:
            mrich.success(f"{reaction_type}: #{check_type} OK")

    else:

        if reac_count - diff != prod_count:
            if debug:
                mrich.error(f"{reaction_type}: #{check_type} FAIL")
            return False

        elif debug:
            mrich.success(f"{reaction_type}: #{check_type} OK")

    return True


def check_atomtype_diff(reaction_type, reactants, product, debug=False):
    """

    :param reaction_type:
    :param reactants:
    :param product:
    :param debug:  (Default value = False)

    """

    check_type = "atomtype"

    # get values
    reac = reactants.atomtype_dict
    prod = product.atomtype_dict

    if debug:
        mrich.var("reactants.atomtype_dict", str(reac))
        mrich.var("product.atomtype_dict", str(prod))

    if "removed" in SUPPORTED_CHEMISTRY[reaction_type]["atomtype"]:
        removal = check_specific_atomtype_diff(
            reaction_type, prod, reac, removal=True, debug=debug
        )

        if not removal:
            return False

    if "added" in SUPPORTED_CHEMISTRY[reaction_type]["atomtype"]:
        addition = check_specific_atomtype_diff(
            reaction_type, prod, reac, removal=False, debug=debug
        )

        if not addition:
            return False

    if debug:
        mrich.success(f"{reaction_type}: atomtypes OK")

    return True


def check_specific_atomtype_diff(reaction_type, prod, reac, removal=False, debug=False):
    """

    :param reaction_type:
    :param prod:
    :param reac:
    :param removal:  (Default value = False)
    :param debug:  (Default value = False)

    """

    if removal:
        add_str = "removed"
    else:
        add_str = "added"

    add_dict = SUPPORTED_CHEMISTRY[reaction_type]["atomtype"][add_str]

    if not add_dict:
        return True

    if debug:
        mrich.var(add_str, str(add_dict))

    for symbol, count in add_dict.items():

        if symbol == "Ha":
            p_count = halogen_count(prod)
            r_count = halogen_count(reac)

        elif symbol == "*":
            assert count == "*", (symbol, count)
            if debug:
                mrich.debug("Allowing wildcard atomtype differences")
            continue

        else:
            p_count = prod[symbol] if symbol in prod else 0
            r_count = reac[symbol] if symbol in reac else 0

        if isinstance(count, str):

            assert count.startswith(">="), (symbol, count)

            count = int(count[2:])

            if removal and r_count - p_count < count:
                if debug:
                    mrich.error(
                        f"{symbol}: {r_count=} - {p_count=} >= {r_count - p_count}"
                    )
                    mrich.error(
                        f"{reaction_type}: atomtype removal {symbol} × {count} FAIL"
                    )
                return False

            elif not removal and p_count - r_count < count:
                if debug:
                    mrich.error(
                        f"{symbol}: {p_count=} - {r_count=} >= {p_count - r_count}"
                    )
                    mrich.error(
                        f"{reaction_type}: atomtype addition {symbol} × {count} FAIL"
                    )
                return False

        else:

            if removal and r_count - p_count != count:
                if debug:
                    mrich.error(
                        f"{symbol}: {r_count=} - {p_count=} = {r_count - p_count}"
                    )
                    mrich.error(
                        f"{reaction_type}: atomtype removal {symbol} × {count} FAIL"
                    )
                return False

            elif not removal and p_count - r_count != count:
                if debug:
                    mrich.error(
                        f"{symbol}: {p_count=} - {r_count=} = {p_count - r_count}"
                    )
                    mrich.error(
                        f"{reaction_type}: atomtype addition {symbol} × {count} FAIL"
                    )
                return False

    return True


def halogen_count(atomtype_dict):
    """

    :param atomtype_dict:

    """
    count = 0
    symbols = ["F", "Cl", "Br", "I"]
    for symbol in symbols:
        if symbol in atomtype_dict:
            count += atomtype_dict[symbol]
    return count


class InvalidChemistryError(Exception):
    """ """

    pass


class UnsupportedChemistryError(Exception):
    """ """

    pass



import mcol
from .price import Price

import mrich


class Quote:
    """Supplier quote for a specific quantity of a :class:`.Compound`.

    .. attention::

            :class:`.Quote` objects should not be created directly. Instead use :meth:`.Compound.get_quotes`.

    """

    _db = None

    def __init__(
        self,
        db: "Database",
        id: int,
        compound: int,
        smiles: str,
        supplier: str,
        catalogue: str,
        entry: str,
        amount: float,
        price: float,
        currency: str,
        purity: float,
        lead_time: int,
        date: str | None = None,
        type: str | None = None,
    ) -> None:

        price = Price(price, currency)

        self._db = db
        self._id = id
        self._compound = compound
        self._smiles = smiles
        self._supplier = supplier
        self._catalogue = catalogue
        self._entry = entry
        self._amount = amount
        self._price = price
        self._purity = purity
        self._lead_time = lead_time
        self._date = date
        self._type = type

        from datetime import datetime

        quote_age = (datetime.today() - datetime.strptime(self.date, "%Y-%m-%d")).days
        # if quote_age > 30:
        # mrich.warning(f'Quote is {quote_age} days old')
        # mrich.warning(self)

    ### FACTORIES

    @classmethod
    def combination(
        cls,
        required_amount: float,
        quotes: list["Quote"],
    ) -> "Quote":
        """Combine a list of quotes into one :class:`.Quote` object.

        * Start with biggest pack
        * Estimate by scaling linearly with unit price

        :param required_amount: amount in mg
        :param quotes: list of quotes to be combined

        """

        biggest_pack = sorted(quotes, key=lambda x: x.amount)[-1]
        unit_price = biggest_pack.price / biggest_pack.amount
        estimated_price = unit_price * required_amount

        self = cls.__new__(cls)
        self.__init__(
            db=biggest_pack.db,
            id=None,
            compound=biggest_pack.compound,
            smiles=biggest_pack.smiles,
            supplier=biggest_pack.supplier,
            catalogue=biggest_pack.catalogue,
            entry=biggest_pack.entry,
            amount=required_amount,
            price=estimated_price.amount,
            currency=biggest_pack.currency,
            purity=biggest_pack.purity,
            lead_time=biggest_pack.lead_time,
            date=biggest_pack.date,
            type=f"estimate from quote={biggest_pack.id}",
        )

        return self

    ### PROPERTIES

    @property
    def entry_str(self) -> str:
        """Unformatted string including the supplier, catalogue (if available), and entry name of the quote"""
        if self.catalogue:
            return f"{self.supplier}:{self.catalogue}:{self.entry}"
        else:
            return f"{self.supplier}:{self.entry}"

    @property
    def db(self) -> "Database":
        """Returns a pointer to the parent database"""
        return self._db

    @property
    def id(self) -> int:
        """Returns the quote's database ID"""
        return self._id

    @property
    def compound(self) -> int:
        """Returns the associated :class:`.Compound`"""
        return self._compound

    @property
    def smiles(self) -> str:
        """Returns the catalogue SMILES string"""
        return self._smiles

    @property
    def supplier(self) -> str:
        """Name of the supplier"""
        return self._supplier

    @property
    def catalogue(self) -> str | None:
        """Name of the catalogue"""
        return self._catalogue

    @property
    def entry(self) -> str:
        """Name/ID of the catalogue entry"""
        return self._entry

    @property
    def amount(self) -> float:
        """Amount in mg"""
        return self._amount

    @property
    def price(self) -> "Price":
        """Price"""
        return self._price

    @property
    def currency(self) -> str:
        """Currency of the associated :class:`.Price` object"""
        return self.price.currency

    @property
    def purity(self) -> float:
        """Purity fraction"""
        return self._purity

    @property
    def lead_time(self) -> float:
        """Lead time in days"""
        return self._lead_time

    @property
    def date(self) -> str:
        """Date the quote was registered to the database"""
        return self._date

    @property
    def type(self) -> str:
        """Description of this quote"""
        return self._type

    @property
    def dict(self) -> dict:
        """Dictionary representation of this quote"""
        return dict(
            id=self.id,
            compound=self.compound,
            smiles=self.smiles,
            supplier=self.supplier,
            catalogue=self.catalogue,
            entry=self.entry,
            amount=self.amount,
            price=self.price,
            purity=self.purity,
            lead_time=self.lead_time,
            date=self.date,
            type=self.type,
        )

    @property
    def currency_symbol(self) -> str:
        """Currency symbol of the associated :class:`.Price`"""
        return self.price.symbol

    ### METHODS

    ### DUNDERS

    def __str__(self):
        """Unformatted string representation"""
        if self.purity:
            purity = f" @ {self.purity:.0%}"
        else:
            purity = ""

        if self.supplier == "Stock":
            return f"C{self.compound} In Stock: {self.amount:}mg{purity}"
        elif self.type:
            return f"C{self.compound} {self.entry_str} {self.amount:}mg{purity} = {self.price:} ({self.lead_time} days) {self.smiles} [{self.type}]"
        else:
            return f"C{self.compound} {self.entry_str} {self.amount:}mg{purity} = {self.price:} ({self.lead_time} days) {self.smiles}"

    def __repr__(self) -> str:
        """ANSI Formatted string representation"""
        return f"{mcol.bold}{mcol.underline}{self}{mcol.unbold}{mcol.ununderline}"

    def __rich__(self) -> str:
        """Rich Formatted string representation"""
        return f"[bold underline]{self}"



__version__ = "0.3.28"

from .animal import HIPPO
from .compound import Compound, Ingredient
from .cset import CompoundSet, CompoundTable, IngredientSet
from .db import Database
from .feature import Feature
from .metadata import MetaData
from .pose import Pose
from .price import Price
from .pset import PoseTable, PoseSet
from .pycule import Quoter
from .quote import Quote
from .reaction import Reaction
from .recipe import Recipe, Route, RouteSet
from .rgen import RandomRecipeGenerator
from .rset import ReactionTable, ReactionSet
from .tags import TagTable, TagSet
from .target import Target
from .scoring import Scorer, CustomAttribute
from .web import ProjectPage
